"""
Module containing custom methods applied to pymatgen structures 
"""

import pymatgen

def get_center_of_mass(structure, indices):
    """
    Return the center of mass of a selection sites of structure selected by indices

    Args:
        structure: pymatgen structure with periodic sites
        indices: list of int, index of selection in structure.sites
    
    Return:
        center_of_mass: cartesian coordinates of the center of mass
    """
    coords = []
    species = []
    reference_site = indices[0] # coord will be computed in a subcell close to this site
    for i in indices:
        species.append(structure.sites[i].species)
        coords_from_ref_to_i = pymatgen.util.coord.pbc_diff(structure.sites[i].frac_coords, structure.sites[reference_site].frac_coords)
        coords.append(structure.lattice.get_cartesian_coords(coords_from_ref_to_i)) # to cartesian coords
    return pymatgen.core.Molecule(species, coords).center_of_mass + structure.sites[reference_site].coords # add reference_coords


#-*- coding: utf-8 -*-

"""
Core functions of coordination search
"""

from ast import Raise
import numpy as np
import pymatgen
from pymatgen.core.structure import Structure
from pymatgen.analysis.molecule_structure_comparator import CovalentRadius
from pymatgen.analysis.graphs import StructureGraph
import networkx as nx
from networkx.algorithms.chains import chain_decomposition
from copy import deepcopy
from collections import Counter
import itertools
from scipy import stats

import logging

import amof.files.path
import amof.structure
import amof.symbols
import amof.files.molsys as ammolsys

# create logger without parameters for this module file that will be incorporated by the main file logging parameters
logger = logging.getLogger(__name__)

class SearchError(Exception):
    """Exception raised when coordination search failed.

    Attributes:
        message -- explanation of the error
        report_search -- dictionary describing the search when the error occurred
    """

    def __init__(self, message, report_search = {}):
        self.message = message
        self.report_search = report_search

class CoordinationSearch(object):
    """
    Classes containing general methods to perform a coordination search
    The actual search is to be launched from a subclass mentioning specific species, etc.

    Attributes:
        fragments: dict of dict;  key i of encompassing dict is fragment of fragnumber i
        atypes, fragtypes, fragnumbers, conn, elems: same structure as molsys.mol object
    """
        
    def __init__(self, struct, neighb_max_distance, dist_margin):
        """Class constructor"""
        self.struct = struct
        self.conn = [[] for i in range(struct.num_sites)]
        self.atypes = ["" for i in range(struct.num_sites)]
        self.fragtypes = ["-1" for i in range(struct.num_sites)]
        self.fragnumbers = [-1 for i in range(struct.num_sites)]
        self.elems = [str(self.struct[i].species)[:-1].lower() for i in range(struct.num_sites)]
        self.fragments = {} 
        self.symbols = amof.symbols.DummySymbols()
        self.all_neighb = self.struct.get_all_neighbors(neighb_max_distance)
        self.dist_margin = dist_margin
        # initialize report_search with useful descriptors of struct for subsequent report analysis
        self.report_search = {"number_of_atoms":self.struct.num_sites}

    def in_fragment(self, index):
        """Return True IFF atom at index 'index' is in a fragment"""
        return self.fragnumbers[index] != -1

    def create_fragment(self, fragtype, indices, fragnumber = 'auto'):
        """
        Create new fragment

        Args:
            fragtype: str
            indices: list of int
            fragnumber: int or str. 
                If int, this fragnumber will be used
                If 'auto' will choose the max value (+1) of existing keys
        """
        if fragnumber == 'auto':
            fragnumber = 0 if len(self.fragments.keys()) == 0 else max(self.fragments.keys()) + 1
        indices = list(set(indices)) # remove duplicates
        fragment = {"fragnumber":fragnumber, "fragtype":fragtype, "indices":indices}
        for i in indices:
            self.fragtypes[i] = fragtype
            self.fragnumbers[i] = fragnumber
        self.fragments[fragnumber] = fragment

    def add_to_fragment(self, fragnumber, indices):
        """
        Add indices to fragment

        Args:
            fragnumber: int
            indices: list of int
        """
        fragtype = self.fragments[fragnumber]['fragtype']
        for i in indices:
            self.fragtypes[i] = fragtype
            self.fragnumbers[i] = fragnumber
        self.fragments[fragnumber]['indices'] = list(set(self.fragments[fragnumber]['indices'] + indices))

    def merge_fragments(self, fragnumber, fragnumber_to_absorb):
        """
        Merge fragments by including fragnumber_to_absorb into fragnumber

        Args:
            fragnumber: int
            fragnumber_to_absorb: int
        """
        self.add_to_fragment(fragnumber, self.fragments[fragnumber_to_absorb]['indices'])
        self.fragments.pop(fragnumber_to_absorb)
        
    
    def change_fragnumber(self, fragnumber_old, fragnumber_new):
        """
        Replace fragnumber if fragnumber_new isn't used

        Args:
            fragnumber_old, fragnumber_new: int
        """
        if fragnumber_new in self.fragments.keys():
            raise SearchError('Cannot change fragnumber: not empty in fragment', self.report_search)
        else:
            fragment = self.fragments.pop(fragnumber_old)
            self.create_fragment(fragment['fragtype'], fragment['indices'], fragnumber = fragnumber_new)

    def clean_fragments(self):
        """
        Perform a sequence of clean-up task in fragments
            - Remove empty fragments
            - Renumber fragments to remove gaps between two frag numbers
        """
        # remove empty fragments
        for fragnumber, fragment in self.fragments.items():
            if len(fragment['indices']) == 0:
                self.fragments.pop(fragnumber)

        # Renumber fragments
        fragnumbers_old = list(self.fragments.keys())
        for i in range(len(fragnumbers_old)):
            if i != fragnumbers_old[i]:
                self.change_fragnumber(fragnumbers_old[i], i)

    def reduce_structure(self):
        """
        Reduce the system by turning fragments into atoms 
            - reduce connectivity
            - change structure by creating a pymatgen site per fragment
        """
        self.make_frag_conn()
        self.symbols.add_names(list(set(self.fragtypes)))
        species = [''] * len(self.fragments)
        coords = [[0. ,0., 0.]] * len(self.fragments)
        for fragnumber, fragment in self.fragments.items():
            species[fragnumber] = self.symbols.get_symbol(fragment['fragtype'])
            coords[fragnumber] = amof.structure.get_center_of_mass(self.struct, fragment['indices'])            
        reduced_struct = Structure(self.struct.lattice, species, coords, coords_are_cartesian = True)

        # compute cutoffs defining nb with reduced structure from frag_conn
        list_of_nb = list(set([tuple(sorted((i, j))) for i in range(len(self.frag_conn)) for j in self.frag_conn[i]])) # no duplicates
        bonds = np.array(['-'.join(sorted([species[i], species[j]])) for (i, j) in list_of_nb])
        bonds_unique = list(set(bonds))
        distances = np.array([reduced_struct.get_distance(i, j) for (i, j) in list_of_nb])
        nb_set_and_cutoff = {}
        for nb_set in bonds_unique:
            nb_set_and_cutoff[nb_set] = np.max(distances[bonds==nb_set])
        self.report_search['nb_set_and_cutoff'] = str(nb_set_and_cutoff)
        
        # check that every fragment B within A-B cutoff of A is nb of A in frag_conn
        # if a fragment B is closer to a fragment C!=A (can be B) than B-A cutoff, then there's no irregularity when checking for B-A. 
        irregular_nb = []
        irregular_nb_offset = []
        nb_list = reduced_struct.get_neighbor_list(max(nb_set_and_cutoff.values()))
        for k in range(len(nb_list)):
            i, j, distance = nb_list[0][k], nb_list[1][k], nb_list[3][k]
            nb_set = '-'.join(sorted([species[i], species[j]]))
            if (nb_set in bonds_unique) and (j not in self.frag_conn[i]) and (distance < nb_set_and_cutoff[nb_set]):
                irregular_nb.append(nb_set)
                irregular_nb_offset.append(nb_set_and_cutoff[nb_set] - distance)
        self.report_search['connectivity_constructible_with_cutoffs'] = (len(irregular_nb) == 0)
        if len(irregular_nb) != 0:
            self.report_search['connectivity_wrongly_inferred_from_cutoffs'] = str(dict(Counter(irregular_nb).items()))
            self.report_search['connectivity_wrong_offsets'] = str(stats.describe(irregular_nb_offset))

        self.report_search['number_of_nodes'] = reduced_struct.num_sites
        self.report_search['symbols'] = str(self.symbols)
        return reduced_struct

    def write_mfpx(self, filename):
        """
        Write an mfpx output file with connectivity info readable by molsys
        """
        mol = ammolsys.DummyMol(self.elems, self.struct.cart_coords, self.struct.lattice.matrix, 
            self.conn, self.atypes, self.fragtypes, self.fragnumbers)
        ammolsys.write_mfpx(mol, filename)

    def make_frag_conn(self):
        """
        Generate a fragment connectivity

        Adapted from molsys
        """
        self.frag_conn = []        # fragment connectivity (indices of fragments)
        self.frag_conn_atoms = []  # atoms making the fragment connectivity (tuples of atom indices: first in fragment, second in other fragment)
        # prepare the atom list for the fragments
        for i in range(len(self.fragments)):
            self.frag_conn.append([])
            self.frag_conn_atoms.append([])
        for i, fragment in self.fragments.items():
            # determine all external bonds of this fragment
            for ia in fragment['indices']:
                for ja in self.conn[ia]:
                    j = self.fragnumbers[ja]
                    if i != j:
                        # this is an external bond
                        self.frag_conn[i].append(j)
                        self.frag_conn_atoms[i].append((ia,ja))

    def get_atype(self, i):
        """Return atype of atom i formatted as in molsys"""
        atype = self.elems[i] + str(len(self.conn[i])) 
        list_of_nn = [self.elems[j] for j in self.conn[i]]
        counts = Counter(list_of_nn)
        list_of_counts = sorted(counts.items(), key= lambda t: (t[0], t[1])) # sort by alphabetical order to have a unique atype per coordination environment
        atype += "_" + ''.join(str(e) for pair in list_of_counts for e in pair)
        return atype

    def update_atypes(self):
        """Update self.atypes"""
        self.atypes = [self.get_atype(i) for i in range(self.struct.num_sites)]

    covalentradius = CovalentRadius().radius # private class variable exclusively used in get_covdist

    def get_covdist(self, i, j): # turn to class at some point to have struct as self. 
        """Return sum of covalent radii of the two atoms referenced by their index i and j"""
        return self.covalentradius[self.elems[i].title()] + self.covalentradius[self.elems[j].title()]

    def add_ABbonds(self, graph, A, B, dist_margin = None):
        """
        Add bonds between species A and B to graph

        Args:
            graph: pymatgen graph
            A, B: str, species written in lower cap (e.g. "h")
            dist_margin: float, margin to consider neighbourings atoms dist_margin further away than their cov distance
        """
        if dist_margin is None:
            dist_margin = self.dist_margin
        for i in range(self.struct.num_sites):
            if self.elems[i] == A:
                for site in self.all_neighb[i]:
                    j = site.index
                    if self.elems[j] == B and site.nn_distance < dist_margin*self.get_covdist(i, j):
                        graph.add_edge(i, j, from_jimage=(0, 0, 0), to_jimage=(0, 0, 0), weight = self.struct.get_distance(i, j), warn_duplicates=False)

    @staticmethod
    def multigraph_to_graph(MG):
        """
        From networkx tutorial, propagate min of weights
        
        Args:
            MG: networkx multigraph
        """
        GG = nx.Graph()
        for n, nbrs in MG.adjacency():
            for nbr, edict in nbrs.items():
                minvalue = min([d['weight'] for d in edict.values()])
                GG.add_edge(n, nbr, weight = minvalue)
        return GG

    @classmethod
    def get_chain_decomposition(cls, graph):
        """
        Return chain decomposition as list of list from pymatgen graph
        
        Args:
            graph: pymatgen graph
        """
        GG = cls.multigraph_to_graph(graph.graph) # pymatgen graph is a multigraph and directed, graph.graph is a nx object
        return list(chain_decomposition(GG)) # direct use of nx function

    @classmethod
    def find_one_cycle_per_node(cls, graph):
        """For each node, if it exists return a cycle found via depth-first traversal in which node is included.

        Args:
            graph: pymatgen graph
            
        Returns:
            cycles_list: list of list of edges"""
        GG = cls.multigraph_to_graph(graph.graph) # pymatgen graph is a multigraph and directed, graph.graph is a nx object
        cycles_list = []
        nodes = list(set(GG.nodes))
        node_in_cycle = {i:False for i in nodes}
        for i in nodes:
            if not node_in_cycle[i]:
                try:
                    cycle = nx.find_cycle(GG, i)
                    cycle_edges = set(itertools.chain.from_iterable(cycle))
                    if i in cycle_edges:
                        for j in cycle_edges:
                            node_in_cycle[j] = True
                        cycles_list.append(cycle)
                except nx.exception.NetworkXNoCycle:
                    pass
        return cycles_list

    @staticmethod
    def are_circularly_identical(arr1, arr2):
        """
        Checks whether two lists of int are circularly identical
        Also works with lists of strings that don't contain spaces ' '

        Code from https://stackoverflow.com/a/26924896/8288189

        Args:
            arr1, arr2: list of ints
        """
        if len(arr1) != len(arr2):
            return False

        str1 = ' '.join(map(str, arr1))
        str2 = ' '.join(map(str, arr2))
        if len(str1) != len(str2):
            return False

        return str1 in str2 + ' ' + str2

    def find_rings(self, graph, including=None, max_depth = None, exit_if_large_cycle = False, 
            pattern=None, target_number_of_rings = None, exit_if_too_many_rings=False, 
            remove_overlapping_rings = False):
        """                
        Find ring structures in the StructureGraph.

        Forked from MoleculeGraph class in pymatgen.analysis.graph
        A ring is defined as a simple cycle in networkx terms
        A simple cycle, or elementary circuit, is a closed path where no node appears twice. 
        Two elementary circuits are distinct if they are not cyclic permutations of each other.

        :param including: list of site indices. If
            including is not None, then find_rings will
            only return those rings including the specified
            sites. By default, this parameter is None, and
            all rings will be returned.
            max_depth: int, maximum cycle size to be found
            exit_if_large_cycle: Bool, if True will raise an Exception if a cycle larger than max_depth is found
            pattern: list of strings representing self.elems to match ('n', 'c', etc.)
            target_number_of_rings: int, number of unique rings
            exit_if_too_many_rings: bool, interupts loop when too many rings found. 
                If False, will raise an error nonetheless but will state how many extra cycles are found
            remove_overlapping_rings: bool, will try to remove overlapping rings 
                Ring A is overlapping if there is ring B such that the intersection of A and B is not empty
                Will be done in a greedy way by removing in the same step every ring that has the largest overlap

        :return: dict {index:cycle}. Each
            entry will be a ring (cycle, in graph theory terms) including the index
            found in the Molecule. If there is no cycle including an index, the
            value will be an empty list.
        """

        # Copies graph such that all edges (u, v) matched by edges (v, u)
        undirected = graph.graph.to_undirected()
        directed = undirected.to_directed()

        cycles_nodes = []
        cycles_edges = []

        # Remove all two-edge cycles
        if max_depth is not None:
            all_cycles = []
            for c in nx.simple_cycles(directed):
                if len(c) > 2 and len(c) <= max_depth:
                    all_cycles.append(c)
                elif exit_if_large_cycle and len(c) > max_depth:
                    raise SearchError('max_depth exceeded in cycle search', self.report_search)
        elif pattern is not None:
            all_cycles = []
            for c in nx.simple_cycles(directed):
                if len(c) == len(pattern):
                    c_pattern = [self.elems[i] for i in c]
                    if self.are_circularly_identical(c_pattern, pattern):
                        all_cycles.append(c)
                if exit_if_too_many_rings and len(all_cycles) > target_number_of_rings * 2:
                    # self.plot_conn_as_graph() 
                    raise SearchError('target_number_of_rings exceeded in pattern cycle search', self.report_search)            
        else:
            all_cycles = [c for c in nx.simple_cycles(directed) if len(c) > 2]

        # Using to_directed() will mean that each cycle always appears twice
        # So, we must also remove duplicates
        unique_sorted = []
        unique_cycles = []
        for cycle in all_cycles:
            if sorted(cycle) not in unique_sorted:
                unique_sorted.append(sorted(cycle))
                unique_cycles.append(cycle)

        if len(unique_cycles) > target_number_of_rings:
            self.report_search['Extra cycles found after initial ring search'] = len(unique_cycles) - target_number_of_rings
        
        # remove rings that are contained in two cycles
        # don't try if not enough rings found to save computation time
        if remove_overlapping_rings == True and len(unique_cycles) > target_number_of_rings:

            def get_cycles_of_atom(unique_cycles):
                """return list of list of cycles containing i for atom i"""
                cycles_of_atom = [[] for i in range(self.struct.num_sites)] 
                for i in range(len(unique_cycles)):
                    for a in unique_cycles[i]:
                        cycles_of_atom[a].append(i)
                return cycles_of_atom

            removed_overlapping_rings = 0
            cycles_of_atom = get_cycles_of_atom(unique_cycles)
            keep_removing_rings = np.max([len(cycles) for cycles in cycles_of_atom]) > 1
            while keep_removing_rings == True:            
                 # each atom is weighted by the number of cycle it's in
                 # each cycle by the sum of the weights of its atoms
                overlap_weight_of_cycle = [np.sum([len(cycles_of_atom[a]) for a in c]) for c in unique_cycles]
                max_weight = np.max(overlap_weight_of_cycle)
                maximum_indices = [i for i, j in enumerate(overlap_weight_of_cycle) if j == max_weight]
                removed_overlapping_rings += len(maximum_indices)
                for index in sorted(maximum_indices, reverse=True):
                    del unique_cycles[index]
                # recontruct cycles_of_atom
                cycles_of_atom = get_cycles_of_atom(unique_cycles)
                keep_removing_rings = np.max([len(cycles) for cycles in cycles_of_atom]) > 1

            self.report_search['Overlapping rings removed'] = removed_overlapping_rings

        if including is None:
            cycles_nodes = unique_cycles
        else:
            for i in including:
                for cycle in unique_cycles:
                    if i in cycle and cycle not in cycles_nodes:
                        cycles_nodes.append(cycle)

        for cycle in cycles_nodes:
            edges = []
            for i, e in enumerate(cycle):
                edges.append((cycle[i - 1], e))
            cycles_edges.append(edges)

        return cycles_edges

    def clean_conn(self):
        """
        Remove duplicates indices in self.conn
        """
        for i in range(len(self.conn)):
            self.conn[i] = list(set(self.conn[i]))

    def plot_conn_as_graph(self, filename = "graph_temp.png"):
        """Create graph with every bond present in self.conn and print it in file"""
        filename = amof.files.path.append_suffix(filename, 'png') # draw_graph_to_file needs a filename extension
        graph = StructureGraph.with_empty_graph(self.struct)
        for i in range(self.struct.num_sites):
            for j in self.conn[i]:
                graph.add_edge(i, j, from_jimage=(0, 0, 0), to_jimage=(0, 0, 0), weight = self.struct.get_distance(i, j), warn_duplicates=False)
        graph.draw_graph_to_file(filename, hide_unconnected_nodes=False, hide_image_edges =False, algo='neato') #algo= neato or fdp
        return graph

    def get_A_Bbonds(self, A, B):
        """
        Return a list containing for each atom A the the number of bonds to a B atom, and -1 if not A atom

        Args:
            A, B: str, species (e.g. "h")
        """
        A_Bbonds = [-1 for i in range(self.struct.num_sites)]
        for i in range(self.struct.num_sites):
            if self.elems[i] == A:
                A_Bbonds[i] = 0
                for j in self.conn[i]:
                    if self.elems[j] == B:
                        A_Bbonds[i] += 1
        return A_Bbonds

    def assign_B_uniquely_to_A_N_coordinated(self, conditionA, conditionB, target_N, 
            use_cov_dist = True, dist_margin=None, report_level = None, report_entry = None, 
            propagate_fragments = False, new_fragments_name = None):
        """
        Assign atoms B to atoms A so that B is only assigned once and A end up being at most target_N coordinated. 

        At each step, the closest pair A-B is added to conn.
        B can only form new bonds with A.
        Allowed to exit if not enough nb available
        
        Args:
            conditionA/B: functions of index i in struct which return a boolean equal to True iff conditionA is verified
            target_N: int, 
                If an A atom already has more than target_N neighbours, will do nothing.
            use_cov_dist: Bool, if True will further restrict the search using cov radii, if False, every nn in all_neighb used
            dist_margin: float
            report_level: str, can be 'full' (report every bond formed) or 'undercoordinated' (only those without enough nn)
            report_entry: str, if supplied, will use it to add as an entry details about the missing nn
            propagate_fragments: Bool, if True will include every atom of frag(B) in fragment of A if frag(A) exists
                If atom B was in a fragment F, F will be removed
            new_fragments_name: str, if not None will create fragments for atoms A not currently in a fragment with fragname "new_fragments_name". 
                To include B atoms in this new fragment, propagate_fragments must be set to True
            
        """
        if dist_margin is None:
            dist_margin = self.dist_margin

        def are_A_Ncoordinated(A_conn, A_enough_nn):
            """Return False as long as the main loop should keep running"""
            for i in range(len(A_conn)):
                if (len(A_conn[i])<target_N) and (A_enough_nn[i]==True):
                    return False
            return True

        # Creation of A_$var variables of length num_of_A_atoms
        A_indices = [] # index of A atom in struct
        A_neighb_indices = [] # indices of nn of A in struct, sorted by increasing distance to A
        A_nn_distances = [] # corresponding distances
        A_conn = [] # contains a temporary copy of conn
        A_enough_nn = [] # True iff A_nn_distances is not empty
        A_new_nb = [] # contains B atoms that will be bounded to A (were not previously in conn)
        
        for i in range(self.struct.num_sites):
            if conditionA(i): 
                A_indices.append(i)
                A_conn.append(deepcopy(self.conn[i])) # starts with preexisting conn
                A_new_nb.append([])
                neighb_set = self.all_neighb[i]
                neighb_set = [n for n in neighb_set if conditionB(n.index)] 
                if use_cov_dist==True:
                    neighb_set = [n for n in neighb_set if n.nn_distance < dist_margin * self.get_covdist(i, n.index)]
                nn_distances = [neighb.nn_distance for neighb in neighb_set]
                sorted_index = np.argsort(nn_distances) # sort nn by distance to nn atom
                A_neighb_indices.append([neighb_set[sorted_index[j]].index for j in range(len(sorted_index))]) # ranked by increasing distance
                A_nn_distances.append([nn_distances[sorted_index[j]] for j in range(len(sorted_index))])
                A_enough_nn.append(not (len(A_conn[-1])<target_N and len(A_nn_distances[-1])==0)) # initialise enough_nn with the current (ie last) element in A_conn and A_nn_distances
        
        while not are_A_Ncoordinated(A_conn, A_enough_nn):
            # take the closest A-B pair and bond them 
            choose_min = [] # candidate of atom A i to be selected
            for i in range(len(A_indices)):
                if len(A_conn[i])>=target_N or A_enough_nn[i]==False:
                    choose_min.append(np.inf) # so that it is not chosen as min
                else:
                    if A_enough_nn[i] == True and len(A_nn_distances[i]) == 0:
                        a = 1
                    choose_min.append(A_nn_distances[i][0]) # the closest nn is the first as A_nn_distances is sorted
            imin = np.argmin(choose_min)
            B_imin = A_neighb_indices[imin][0]
            A_conn[imin].append(B_imin) # add to A_conn
            A_new_nb[imin].append(B_imin)
            for i in range(len(A_indices)): # remove B_imin as potential candidate for every A atom
                while B_imin in A_neighb_indices[i]:
                    A_nn_distances[i].pop(A_neighb_indices[i].index(B_imin))
                    A_neighb_indices[i].remove(B_imin)
            # flag A atoms missing nn
            for i in range(len(A_indices)):
                if A_enough_nn[i]==True and len(A_nn_distances[i])==0:
                    A_enough_nn[i]=False
        
        # add A_conn to self.conn
        for i in range(len(A_indices)): 
            A = A_indices[i]
            self.conn[A]=A_conn[i]
            for n in A_conn[i]:
                if A not in self.conn[n]: # if conn wasn't empty at the beginning
                    self.conn[n].append(A)
        

        if report_level=='full':
            list_of_atypes = [self.get_atype(i) for i in range(self.struct.num_sites) if conditionA(i)]
            self.report_search[report_entry] = Counter(list_of_atypes).most_common() # return list sorted by decreasing number of occurances
            if self.report_search[report_entry] != []:
                logger.debug("%s: %s", report_entry, self.report_search[report_entry])

        if report_level=='undercoordinated':
            list_of_atypes = [self.get_atype(A_indices[i]) for i in range(len(A_indices)) if (len(A_conn[i])!=target_N)]
            self.report_search[report_entry] = Counter(list_of_atypes).most_common() # return list sorted by decreasing number of occurances
            if self.report_search[report_entry] != []:
                logger.debug("%s: %s", report_entry, self.report_search[report_entry])

        # Create new fragments for A atoms
        if new_fragments_name is not None:
            for a in A_indices:
                if not self.in_fragment(a):
                    self.create_fragment(new_fragments_name, [a])
        
        # Propagate fragments
        if propagate_fragments:
            for i in range(len(A_indices)): 
                a = A_indices[i]
                if self.in_fragment(a):
                    for b in A_new_nb[i]:
                        if not self.in_fragment(b):
                            self.add_to_fragment(self.fragnumbers[a], [b])
                        else:
                            self.merge_fragments(self.fragnumbers[a], self.fragnumbers[b])

    def get_neighb_cov_dist(self, i, dist_margin=None):
        """Return list of nb of atom i using a cutoff based on covalent radii * dist_margin"""
        if dist_margin is None:
            dist_margin = self.dist_margin
        return [n for n in self.all_neighb[i] if n.nn_distance < dist_margin * self.get_covdist(i, n.index)]

    def find_N_closest_cov_dist(self, conditionA, conditionB, target_N, dist_margin = None, 
        report_level = None, report_entry = None, propagate_fragments = False, new_fragments_name = None):
        """
        Find target_N nearest neighbours respecting conditionB to conditionA atoms
        The search for each A is independant, as a result the same B atom can be bonded to several A atoms

        Args:
            conditionA/B: functions of index i in struct which return a boolean equal to True iff conditionA is verified
            target_N: int
            dist_margin: float
            report_level: str, can be 'full' (report every bond formed) or 'undercoordinated' (only those without enough nn)
            report_entry: str, if supplied, will use it to add as an entry details about the missing nn
            propagate_fragments: Bool or str if True will include every atom of frag(B) in fragment of A if frag(A) exists
                If atom B was in a fragment F, F will be removed
                If 'reverse' and target_N = 1, will propagate from B to A
            new_fragments_name: str, if not None will create fragments for atoms A not currently in a fragment with fragname "new_fragments_name". 
                To include B atoms in this new fragment, propagate_fragments must be set to True
        """
        if dist_margin is None:
            dist_margin = self.dist_margin

        list_of_undercoordinated = []

        A_indices = [] # index of A atom in struct
        A_new_nb = [] # contains B atoms that will be bounded to A (were not previously in conn)

        for i in range(self.struct.num_sites):
            if conditionA(i):
                A_indices.append(i)
                new_nb = []
                neighb_set = self.get_neighb_cov_dist(i, dist_margin)
                neighb_set = [s for s in neighb_set if conditionB(s.index)]
                if len(neighb_set)<target_N:
                    logger.debug("not enough nn for %s: %s instead of %s",i, len(neighb_set), target_N)
                    for s in neighb_set:
                        logger.debug("%s %s distance: %s conn: %s", self.elems[s.index], s.index, s.nn_distance, str([(self.elems[j], j) for j in self.conn[s.index]]))
                    list_of_undercoordinated.append(i)
                # take the closest as nn
                nn_distances = [neighb.nn_distance for neighb in neighb_set]
                sorted_index = np.argsort(nn_distances)
                for j in range(min(target_N, len(neighb_set))):
                    nn=neighb_set[sorted_index[j]]
                    new_nb.append(nn.index)
                    self.conn[i].append(nn.index)
                    self.conn[nn.index].append(i)
                A_new_nb.append(new_nb)
        
        if report_level=='full':
            list_of_atypes = [self.get_atype(i) for i in range(self.struct.num_sites) if conditionA(i)]
            self.report_search[report_entry] = Counter(list_of_atypes).most_common() # return list sorted by decreasing number of occurances
            if self.report_search[report_entry] != []:
                logger.debug("%s: %s", report_entry, self.report_search[report_entry])

        if report_level=='undercoordinated':
            list_of_atypes = [self.get_atype(i) for i in list_of_undercoordinated]
            self.report_search[report_entry] = Counter(list_of_atypes).most_common() # return list sorted by decreasing number of occurances
            if self.report_search[report_entry] != []:
                logger.debug("%s: %s", report_entry, self.report_search[report_entry])
        
        # Create new fragments for A atoms
        if new_fragments_name is not None:
            for a in A_indices:
                if not self.in_fragment(a):
                    self.create_fragment(new_fragments_name, [a])

        # Propagate fragments
        if propagate_fragments == True:
            for i in range(len(A_indices)): 
                a = A_indices[i]
                if self.in_fragment(a):
                    for b in A_new_nb[i]:
                        if not self.in_fragment(b):
                            self.add_to_fragment(self.fragnumbers[a], [b])
                        else:
                            self.merge_fragments(self.fragnumbers[a], self.fragnumbers[b])
        elif propagate_fragments == 'reverse':
            if target_N != 1:
                raise SearchError("Propagation ambiguous: Tried to propagate fragment from B to A with target_N not equal to 1", self.report_search) 
            else:
                for i in range(len(A_indices)): 
                    a = A_indices[i]
                    if len(A_new_nb[i]) == 1:
                        b = A_new_nb[i][0]
                        if not self.in_fragment(a):
                            self.add_to_fragment(self.fragnumbers[b], [a])
                        else:
                            self.merge_fragments(self.fragnumbers[b], self.fragnumbers[a])
                        

class NotImplementedSearch(CoordinationSearch):
    """
    Dummy coordination search class for MOFs for which it is not implemented
    """

    def __init__(self, mof = None):
        """
        Constructor of ZifSearch

        Args:
            mof: str, name of mof for which search is not implemented
        """
        self.report_search = {} 
        if mof is not None:
            self.report_search['mof'] = mof
        raise SearchError('Structure search not implemented', self.report_search)
 






"""
Module containing everything related to trajectories: 
- wrapper for ase trajectory module
- constructor from different files used
- ReducedTrajectory class
"""

# import ase.io.trajectory
import ase.io
from ase.geometry.geometry import wrap_positions

import logging
import bisect
import numpy as np
import gzip
import shutil
import tempfile
import pandas as pd
import pathlib

import amof.atom
import amof.files.path as ampath
import amof.symbols

logger = logging.getLogger(__name__)

class Trajectory(object):
    """
    Wrapper for ase trajectory module
    self.traj: List of frames, each frame is an ase atom object
    """
    def __init__(self):
        """default constructor"""
        self.traj = []
        # self.struct = struct

    @classmethod
    def from_traj(cls, filename, index = None, format = None, unzip = False):
        """
        constructor of trajectory class using ase.io.read

        Args:
            index: index using ase format: 'first_frame:last_frame:step' or slice(first_frame,last_frame,step)
            format: str, Used to specify the file-format. If not given, the file-format will be guessed by the filetype function.
            unzip: if True will unzip file as temporary file before reading it with ase, if False will let ase handle the decompression
        """
        logger.info("Read trajectory %s", filename)
        trajectory_class = cls() # initialize class
        # cls.traj = ase.io.trajectory(filename, mode='r')
        if unzip:
            logger.info("Unzip trajectory file")
            with tempfile.NamedTemporaryFile(buffering=0) as tmp: # beffering messes with ase read in some way, desactivating it
                with gzip.open(filename, 'rb') as f_in:
                    shutil.copyfileobj(f_in, tmp)
                logger.info("Read trajectory with ase")
                trajectory_class.traj = ase.io.read(tmp.name, index, format)
        else:
            logger.info("Read trajectory with ase")
            trajectory_class.traj = ase.io.read(filename, index, format)
        return trajectory_class 

    @classmethod
    def from_lammps_data(cls, filename, atom_style):
        """
        constructor of trajectory class from lammps data file 'filename'
        atom_style: string representing lammps atom_style (e.g. 'charge')
        """
        trajectory_class = cls() # initialize class
        atoms = ase.io.read(filename, format = 'lammps-data', style=atom_style)
        myList = ase.data.atomic_masses            
        atomic_numbers = [cls.get_index_closest(myList, myNumber) for myNumber in atoms.get_masses()]            
        atoms.set_atomic_numbers(atomic_numbers)
        trajectory_class.traj = [atoms]
        return trajectory_class 

    @staticmethod
    def get_index_closest(myList, myNumber):
        """
        Assumes myList is sorted. Returns closest value to myNumber.

        If two numbers are equally close, return the smallest number.
        """
        pos = bisect.bisect_left(myList, myNumber)
        # pos = np.searchsorted(myList, myNumber) # twice as slow in tried example
        if pos == 0:
            return myList[0]
        if pos == len(myList):
            return myList[-1]
        before = myList[pos - 1]
        after = myList[pos]
        if after - myNumber < myNumber - before:
            return pos
        else:
            return pos - 1

    def set_cell(self, cell, set_pbc = True, fit_size = True):
        """
        Args: 
            cell: cell vector of same length than self.traj in valid ase cell format
            set_pbc: wether to set_pbc=True to every frame of traj
            fit_size = if True will limit array size to the smallest if cell and traj to avoid raising an error
        """
        if fit_size and len(self.traj)!=len(cell):
            logger.warning("Mismatch in file sizes; traj: %s vs cell: %s", len(self.traj), len(cell))
            if len(self.traj) > len(cell):
                logger.warning("Reducing traj length")
                self.traj = self.traj[0:len(cell)]
            else:
                logger.warning("reducing cell length")
                cell = cell[0:len(self.traj)]
        for i in range(len(self.traj)):
            self.traj[i].set_cell(cell[i])
            if set_pbc:
                self.traj[i].set_pbc(True)

    def get_traj(self):
        return self.traj


class ReducedTrajectory(object):
    """
    Class representing a reduced trajectory

    Attributes:
        trajectory: ase trajectory object containing the positions of the reduced trajectory
        report_search: pandas dataframe with information per Step of trajectory
        symbols: amof DummySymbols object
    """
    def __init__(self, trajectory = [], 
            report_search = pd.DataFrame({"Step": np.empty([0])}), 
            symbols = amof.symbols.DummySymbols()):
        """Default constructor from attributes
        Mock reducedtrajectory with no frame if no argument supplied"""
        self.trajectory = trajectory
        self.report_search = report_search
        self.symbols = symbols

    @classmethod
    def from_file(cls, filename, sampling = 1, load_trajectory = True):
        """
        Constructor of class from files

        Args: 
            filename: str or path to files without the final suffixes (ie no '.xyz' or '.symbols')
            sampling: int, will sample the trajectory with this rate choosing frames that are in_reduced_trajectory
        """
        if load_trajectory == True:
            logger.info("Read reduced trajectory %s", pathlib.Path(filename).name)
            trajectory = ase.io.read(ampath.append_suffix(filename, 'xyz'), ':', 'xyz')
        else:
            trajectory = []
        report_search = pd.read_csv(ampath.append_suffix(filename, 'report_search.csv'), index_col=0)
        symbols = amof.symbols.DummySymbols.from_file(filename)
        cn_class = cls(trajectory, report_search, symbols) # initialize class
        if sampling != 1:
            cn_class.sample(sampling)
        return cn_class # return class as it is a constructor

    def write_to_file(self, filename):
        """
        Args: 
            filename: str or path to files without the final suffixes (ie no '.xyz' or '.symbols')
        """
        self.report_search.to_csv(ampath.append_suffix(filename, 'report_search.csv'))
        ase.io.write(ampath.append_suffix(filename, 'xyz'), self.trajectory)
        self.symbols.write_to_file(filename)

    def sample(self, sampling):
        """
        Sample trajectory choosing frames that are in_reduced_trajectory

        sampling refers to the frames indexed by integers (0,1,2..) and to Steps
        First time it is used, it is relative to the not reduced trajectory
        Then if the reduced traj has already been sampled, then the rate is relative to the sampled reduced trajectory

        Args: 
            sampling: int, sampling rate relative to not reduced trajectory
        """        
        if len(self.report_search) != 0:
            rs_traj = self.report_search[self.report_search['in_reduced_trajectory']==True]
            in_traj_sampling = round(sampling * len(rs_traj) / len(self.report_search)) # can be zero if few frames in reduced traj
            if in_traj_sampling != 0: 
                self.trajectory = self.trajectory[::in_traj_sampling]  
                self.report_search = rs_traj[::in_traj_sampling]

def read_lammps_data(filename, atom_style):
    """
    constructor of trajectory class from lammps data file 'filename'
    atom_style: string representing lammps atom_style (e.g. 'charge')
    """
    return Trajectory.from_lammps_data(filename, atom_style).get_traj()

def read_lammps_traj(path_to_xyz, index = None, cell = None):
    """
    Args: 
        index: index using ase format: 'first_frame:last_frame:step' or slice(first_frame,last_frame,step)
        cell: cell vector of same length than self.traj, if None doesn't set cell
        
    Returns:
        traj: ase.trajectory object
    """
    Traj = Trajectory.from_traj(path_to_xyz, index, format = 'xyz')
    if cell is not None:
        Traj.set_cell(cell, set_pbc = True)
    return Traj.get_traj()


def read_cp2k_traj(path_to_xyz, path_to_cell, index = None, unzip_xyz = False):
    """
    Args: 
        index: index using slice ase format: slice(first_frame,last_frame,step). 
        'first_frame:last_frame:step' isn't supported
        
    Returns:
        traj: ase.trajectory object
    """
    Traj = Trajectory.from_traj(path_to_xyz, index, format = 'xyz', unzip = unzip_xyz)
    cell = np.genfromtxt(path_to_cell)
    if len(cell.shape) == 1: # corresponds to single frame in traj
        cell = cell[2:-1] 
        cell = cell[index]
        cell = np.array([cell.reshape(3,3)]) # reshape in 3*3 matrix cell format
    else:
        cell = cell[:,2:-1] 
        cell = cell[index]
        cell = np.array([c.reshape(3,3) for c in cell]) # reshape in 3*3 matrix cell format
    Traj.set_cell(cell, set_pbc = True)
    return Traj.get_traj()


def apply_to_traj(trajectory, function, how):
    """apply function to every atom of trajectory and aggregate using how"""
    if how == 'mean':
        return np.mean([function(atom) for atom in trajectory])

def get_density(trajectory, how = 'mean'):
    """return density of ase trajectory object"""
    return apply_to_traj(trajectory, amof.atom.get_density, how)

def get_number_density(trajectory, how = 'mean'):
    """return number density of ase trajectory object"""
    return apply_to_traj(trajectory, amof.atom.get_number_density, how)

def construct_step(**kwargs):
    """
    Construct step from various constructors

    Args:
        delta_Step: int, number of simulation steps between two frames
        first_frame: int, first step
        last_frame: int, first step
        number_of_frames: int, number of frames
        step: slice object or numpy array
    
    Return:
        step: numpy array containing steps
    """
    delta_Step = kwargs.get('delta_Step', None)
    first_frame = kwargs.get('first_frame', None)
    last_frame = kwargs.get('last_frame', None)
    number_of_frames = kwargs.get('number_of_frames', None)
    step = kwargs.get('step', None)

    try:
        if step is not None:
            if isinstance(step, slice):
                return np.array(list(range(step.start or 0, step.stop, step.step or 1)))
            else:
                return np.array(step)
        elif delta_Step is not None:
            if first_frame is not None and last_frame is not None:
                return np.arange(first_frame, last_frame, delta_Step)
            elif number_of_frames is not None: 
                if first_frame is None and last_frame is not None:
                    first_frame = last_frame - number_of_frames * delta_Step
                if first_frame is not None:
                    return np.arange(first_frame, first_frame + number_of_frames * delta_Step, delta_Step)
        elif number_of_frames is not None:
            if first_frame is not None and last_frame is not None:
                return np.linspace(first_frame, last_frame, number_of_frames)
    except:
        logger.exception("Cannot construct step from provided args")       
        raise ValueError  

def get_delta_pos(pos, cell):
    """
    Converts position coordinates of trajectory into displacements between consecutive frames
    Each displacement will be made to fit in the unit cell around (0,0,0). 
    In other words it will be as small as possible
    The first frame contains the initial positions
    For each frame k the original positions can be retrieved by summing all the displacements from 0 to k

    Args:
        pos: list, a trajectory of atomic positions
        cell: list, a traj of cell values (can be ase cell format)

    Return:
        delta_pos: list, a trajectory of atomic positions
    """
    delta_pos = [pos[0]]
    for k in range(len(pos)-1):
        delta_pos.append(wrap_positions(pos[k+1]-pos[k], cell[k], center=(0., 0., 0.)))
    return delta_pos


"""
Module containing rdf related methods
"""

import os
os.environ["OMP_NUM_THREADS"] = "1" # export OMP_NUM_THREADS=4

import ase
import ase.data
import asap3.analysis.rdf

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.integrate
import scipy.interpolate
import joblib

import logging

import amof.trajectory
import amof.files.path
import amof.atom as amatom

# create logger without parameters for this module file that will be incorporated by the main file logging parameters
logger = logging.getLogger(__name__)

class Rdf(object):
    """
    Main class for rdf
    """

    def __init__(self):
        """default constructor"""
        self.data = pd.DataFrame({"r": np.empty([0])})

    @classmethod
    def from_trajectory(cls, trajectory, dr = 0.01, rmax = 'half_cell'):
        """
        Constructor of rdf class 
        
        Args:
            trajectory: an ase trajectory object
            dr, rmax: floats in Angstrom
                If rmax is set to 'half_cell', then half of the minimum dimension of the cell is used 
                to ensure no atom is taken into account twice for a given atom 
                (computation is possible beyond this threeshold up to the min cell size)
        """
        rdf_class = cls() # initialize class
        rdf_class.compute_rdf(trajectory, dr, rmax)
        return rdf_class # return class as it is a constructor

    @classmethod
    def from_rdf(cls, *args):
        logger.exception('from_rdf is deprecated, use from_file instead')

    @classmethod
    def from_file(cls, path_to_rdf):
        """
        constructor of rdf class from rdf file
        """
        rdf_class = cls() # initialize class
        rdf_class.read_rdf_file(path_to_rdf)
        return rdf_class # return class as it is a constructor


    def compute_rdf(self, trajectory, dr, rmax):
        """
        compute rdf from ase trajectory object
        """
        atomic_numbers_unique = list(set(trajectory[0].get_atomic_numbers()))
        N_species = len(atomic_numbers_unique) # number of different chemical species

        rmax_half_cell = np.min([a for t in trajectory for a in t.get_cell_lengths_and_angles()[0:3]]) / 2
        if  rmax == 'half_cell':        # default option
            rmax = rmax_half_cell
        elif rmax > rmax_half_cell:
            logger.info("Specified rmax %s is larger than half cell; will use half_cell rmax", rmax)
            rmax = rmax_half_cell

        logger.info("Start computing rdf for %s frames with dr = %s and rmax = %s", len(trajectory), dr, rmax)
        bins = int(rmax // dr)
        r = np.arange(bins) * dr        
        self.data = pd.DataFrame({"r": r})

        # Code from the asap3 manual for a trajectory
        RDFobj = None
        for atoms in trajectory:
            if RDFobj is None:
                RDFobj = asap3.analysis.rdf.RadialDistributionFunction(atoms, rmax, bins)
            else:
                RDFobj.atoms = atoms  # Fool RDFobj to use the new atoms
            RDFobj.update()           # Collect data
        
        # Total RDF        
        rdf = RDFobj.get_rdf(groups=0)        
        self.data["X-X"] = rdf

        # Partial RDFs               
        elements = [[(x, y) for y in atomic_numbers_unique] for x in atomic_numbers_unique] # cartesian product of very couple of  species

        # change to np.like 
        partial_rdf = [[0 for y in atomic_numbers_unique] for x in atomic_numbers_unique] # need to have same structure as elements but any content is fine as it will be replaced

        for i in range(N_species):
            for j in range(N_species):
                xx = elements[i][j]
                xx_str = ase.data.chemical_symbols[xx[0]] + "-" + ase.data.chemical_symbols[xx[1]]
                partial_rdf[i][j] = RDFobj.get_rdf(elements=xx,groups=0)
                self.data[xx_str] = partial_rdf[i][j]    
        for i in range(N_species):
            xx = elements[i][i]
            xx_str = ase.data.chemical_symbols[xx[0]] + "-" + ase.data.chemical_symbols[xx[1]]
            self.data[ase.data.chemical_symbols[xx[0]] + "-X"] = sum([partial_rdf[i][j] for j in range(N_species)])   

    def write_to_file(self, filename):
        filename = amof.files.path.append_suffix(filename, 'rdf')
        self.data.to_feather(filename)

    def read_rdf_file(self, path_to_data):
        path_to_data = amof.files.path.append_suffix(path_to_data, 'rdf')
        self.data = pd.read_feather(path_to_data)

    def get_coordination_number(self, nn_set, cutoff, density):
        """
        return coordination number
        nn_set: str indicating pair of neighbours
        cutoff: float, in Angstrom
        density: float, no units
        """
        return get_coordination_number(self.data['r'], self.data[nn_set], cutoff, density)



class CoordinationNumber(object):
    """
    Class to compute CoordinationNumber from RDF
    
    Subjected to numerical errors in the integration step
    Best to use amof.cn.CoordinationNumber
    """

    def __init__(self):
        """default constructor"""
        logger.warning('Compute CoordinationNumber from RDF, best to use amof.cn.CoordinationNumber')
        self.data = pd.DataFrame({"Step": np.empty([0])})

    @classmethod
    def from_trajectory(cls, trajectory, nb_set_and_cutoff, delta_Step = 1, first_frame = 0, dr = 0.0001, parallel = False):
        """
        constructor of rdf class from an ase trajectory object
        Args:
            nb_set_and_cutoff: dict, keys are str indicating pair of neighbours, 
                values are cutoffs float, in Angstrom
            dr: float, in Angstrom
        """
        cn_class = cls() # initialize class
        step = amof.trajectory.construct_step(delta_Step=delta_Step, first_frame = first_frame, number_of_frames = len(trajectory))
        cn_class.compute_cn(trajectory, nb_set_and_cutoff, step, dr, parallel)
        return cn_class # return class as it is a constructor

    def compute_cn(self, trajectory, nb_set_and_cutoff, step, dr, parallel):
        """
        compute coordination from ase trajectory object
        """
        atomic_numbers_unique = list(set(trajectory[0].get_atomic_numbers()))
        N_species = len(atomic_numbers_unique) # number of different chemical species

        rmax = np.max(list(nb_set_and_cutoff.values()))

        logger.info("Start computing coordination number for %s frames with dr = %s and rmax = %s", len(trajectory), dr, rmax)
        bins = int(rmax // dr)
        r = np.arange(bins) * dr        

        def compute_cn_for_frame(atom, step):
            """
            compute coordination for ase atom object
            """
            # atoms = trajectory[i]
            dic = {'Step': step}
            RDFobj = asap3.analysis.rdf.RadialDistributionFunction(atom, rmax, bins)
            density = amatom.get_number_density(atom)
            for nn_set, cutoff in nb_set_and_cutoff.items():
                xx = tuple(ase.data.atomic_numbers[i] for i in nn_set.split('-'))
                rdf = RDFobj.get_rdf(elements=xx, groups=0)
                dic[nn_set] = get_coordination_number(r, rdf, cutoff, density)            
            return dic

        if parallel == False:
            list_of_dict = [compute_cn_for_frame(trajectory[i], step[i]) for i in range(len(trajectory))]
        else:
            logger.warning("Parallel mode for coordination number very slow, best to use serial")
            num_cores = parallel if type(parallel) == int else max(joblib.cpu_count() // 2 - 2, 2) # heuristic for 40cores Xeon cpus
            list_of_dict = joblib.Parallel(n_jobs=num_cores)(joblib.delayed(compute_cn_for_frame)(trajectory[i], step[i]) for i in range(len(trajectory)))

        self.data = pd.DataFrame(list_of_dict)

    @classmethod
    def from_file(cls, filename):
        """
        constructor of pore class from msd file
        """
        cn_class = cls() # initialize class
        cn_class.read_cn_file(filename)
        return cn_class # return class as it is a constructor

    def read_cn_file(self, filename):
        """path_to_data: where the cn object is"""
        filename = amof.files.path.append_suffix(filename, 'cn')
        self.data = pd.read_feather(filename)

    def write_to_file(self, filename):
        filename = amof.files.path.append_suffix(filename, 'cn')
        self.data.to_feather(filename)

def get_coordination_number(r, rdf, cutoff, density):
    """
    return coordination number
    x, rdf: arrays of same size
    cutoff: float, in Angstrom
    density: float, number density of the entire system (counting every species) in Angstrom^-3
    """
    mask = (r > 0) & (r < cutoff)
    r = r[mask]
    rdf = rdf[mask]
    integral = scipy.integrate.simps(rdf * (r**2), r)
    return 4 * np.pi * density * integral


class RdfPlotter(object):
    def __init__(self):
        """default constructor"""
        self.multiple_rdf_data = {}
        # self.struct = struct

    def add_rdf(self, path_to_rdf, rdf_name = None):
        if rdf_name is None:
            rdf_name = path_to_rdf
        self.multiple_rdf_data[rdf_name] = Rdf.from_rdf(path_to_rdf).data

    # must clean single_frame
    @classmethod
    def from_multiple_rdf(cls, list_of_path_to_rdf, list_of_rdf_name = None):
        """
        constructor of rdf class from rdf file
        """
        if list_of_rdf_name is None:
            list_of_rdf_name = list_of_path_to_rdf
        rdf_plotter_class = cls() # initialize class
        for i in range(len(list_of_path_to_rdf)):
            rdf_plotter_class.add_rdf(list_of_path_to_rdf[i], list_of_rdf_name[i])
        return rdf_plotter_class # return class as it is a constructor

    def plot(self, nn_set, path_to_plot = None, xlim = None):
        """
        plot the rdf of a given neighbour nn_set (e.g. "X-X" or "Zn-Zn") for every rdf_data in multiple_rdf_data
        """
        for rdf_name, rdf_data in self.multiple_rdf_data.items():
            plt.plot(rdf_data['r'], rdf_data[nn_set], label = rdf_name,  alpha=0.9, linewidth = 1)
        plt.legend()
        plt.xlabel("$r$ (${\AA}$)")
        plt.ylabel("$g(r)$")
        if xlim is not None:
            plt.xlim(xlim[0],xlim[-1])
        plt.title(nn_set)
        if path_to_plot is not None:
            plt.savefig(path_to_plot + ".png", dpi = 300)
        plt.show()    
   




"""
Module containing methods to output files readable by molsys:
    - mfpx

Forked and adapted from molsys/fileIO/mfpx.py

Molsys is part of MOFplus/cmc-tools, licensed under the MIT License
Copyright (c) 2021 MOF+ (Computational Materials Chemistry group, Ruhr-Universität Bochum)
https://github.com/MOFplus/cmc-tools
"""


import numpy
import logging

import amof.files.path as ampath

"""
These functions implement the reading and writing of mfpx files

for the handling of metadata of topologies/embedding the system of keywords is largely extended.
Note: the code still reads old mfpx file of type top correctly but it writes only the new format.

"""

logger = logging.getLogger(__name__)

class DummyMol(object):
    """
    DummyMol object with all the necessary attributes
    Mainly here to provide info on mol arg for write_mfpx
    """
    def __init__(self, elems, xyz, cell, conn, atypes, fragtypes, fragnumbers):
        """default constructor"""
        self.cell = cell
        self.fragtypes = fragtypes
        self.fragnumbers = fragnumbers
        self.elems = elems
        self.xyz = xyz
        self.conn = conn
        self.natoms = len(elems)
        self.atypes = atypes

def write_mfpx(mol, filename):
    """
    Routine, which writes an mfpx file
    :Parameters:
        -mol   (obj) : DummyMol or class with similar attributes
        -filename (str) : filename
    """
    filename = ampath.append_suffix(filename, 'mfpx')
    f = open(filename, 'w')
    ### write check ###
    try:
        f.write ### do nothing
    except AttributeError:
        raise IOError("%s is not writable" % f)
    ### write func ###
    ftype = 'xyz'
    f.write('# type %s\n' % ftype)
    f.write('# cellvect %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f\n' %\
        tuple(mol.cell.ravel()))
    f.write('%i\n' % mol.natoms)
    write_body(f,mol)
    f.close()

def write_body(f, mol):
    """
    Routine, which writes the body of a txyz or a mfpx file
    :Parameters:
        -f      (obj)  : fileobject
        -mol    (obj)  : DummyMol or class with similar attributes
    """
    fragtypes   = mol.fragtypes
    fragnumbers = mol.fragnumbers
    elems       = mol.elems
    xyz         = mol.xyz
    cnct        = mol.conn
    natoms      = mol.natoms
    atypes      = mol.atypes

    xyzl = xyz.tolist()
    for i in range(natoms):
        line = ("%3d %-3s" + 3*"%12.6f" + "   %-24s") % \
            (i+1, elems[i], xyzl[i][0], xyzl[i][1], xyzl[i][2], atypes[i])
        line += ("%-16s %5d ") % (fragtypes[i], fragnumbers[i])
        conn = (numpy.array(cnct[i])+1).tolist()
        if len(conn) != 0:
            line += (len(conn)*"%7d ") % tuple(conn)
        f.write("%s \n" % line)



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.13.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Compute structural properties
#
# Using the example of a single frame trajectory of a ZIF-4 unit cell, we'll explore the different analyses made possible by aMOF.

# %%
# Import hvplot for easy plotting
import hvplot.xarray # noqa
import hvplot.pandas # noqa
import numpy as np

# %% [markdown]
# #### Import trajectory
# Read with ase

# %%
import ase.io
atom = ase.io.read('files/ZIF-4.xyz')
traj = [atom] # single atom trajectory

# %%
atom

# %% [markdown]
# ### RDF (Radial Distribution Function)
# Each analysis is contained in a module, here `rdf`

# %%
import amof.rdf as amrdf
help(amrdf)

# %% [markdown]
# The module contains a main class used to compute the property, here 'Rdf'.

# %%
help(amrdf.Rdf)

# %% [markdown]
# This class contains a constructor to compute it from an ase trajectory: `.from_trajectory()`

# %%
rdf_object = amrdf.Rdf.from_trajectory(traj)

# %% [markdown]
# The information is always stored in an instance attribute called `.data`, which can take various forms depending on the data stored. It usually is a [pandas dataframe](https://pandas.pydata.org/) or an [xarray DataArray](https://docs.xarray.dev/en/stable/).
#
# For the RDF it's a dataframe with the position as index, and name of the atom pair as column:

# %%
rdf_object.data

# %%
rdf_object.data.hvplot(x='r', y = ['C-H', 'Zn-N'])

# %% [markdown]
# This can be stored to a file with the generic `write_to_file` method and read again from this file

# %%
rdf_object.write_to_file('filename.rdf')
new_rdf_object = amrdf.Rdf.from_file('filename.rdf')

# Do we have the same values?
np.allclose(new_rdf_object.data, rdf_object.data)

# %% [markdown]
# ### Bond-Angle distribution
# module `bad`

# %%
import amof.bad as ambad

# %%
bad_object = ambad.Bad.from_trajectory(traj, {'Zn-N': 2.5}) 

# %%
bad_object.data.hvplot(x='theta', y ="N-Zn-N", xlim = (100,120))

# %% [markdown]
# ### Coordination number
# module `cn`

# %%
import amof.cn as amcn
cn_object = amcn.CoordinationNumber.from_trajectory(traj, {'Zn-N': 2.5})
cn_object.data

# %% [markdown]
# ### Mean Squared Displacement (MSD)
# module `msd`
#
# Create a mock trajectory with random noise just to have several frames to show the MSD computations.

# %%
from copy import deepcopy
mock_traj = [deepcopy(atom)]
for i in range(10):
    atom.rattle(0.5)
    mock_traj.append(deepcopy(atom))

# %%
import amof.msd as ammsd
msd = ammsd.WindowMsd.from_trajectory(mock_traj, delta_time = 1, timestep = 1)        

# %%
msd.data.hvplot(y = ['X','Zn','C'])

# %% [markdown]
# ### Pore analysis
# module `pore`
#
# Calls Zeo++

# %%
import amof.pore as ampore
pore = ampore.Pore.from_trajectory(traj) # takes a few seconds to run

# %%
pore.data

# %% [markdown]
# ### Ring analysis
# module `ring`
#
# Here we show an example where we set-up the search only to look for imidazolates (ie cycles of C-N-C-N-C atoms).
# It takes a few (~5) minutes to run.

# %%
import amof.ring as amring
ring = amring.Ring.from_trajectory(traj, {'C-N': 1.728, 'C-C': 1.752}, max_search_depth = 6) # takes a few minutes to run

# %%
ring.data.ring.to_series()

# %%



"""
Module containing coordination number related methods
"""

import os
os.environ["OMP_NUM_THREADS"] = "1" # export OMP_NUM_THREADS=4

import ase
import ase.data

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import joblib

import logging

import amof.trajectory
import amof.files.path
import amof.atom as amatom

# create logger without parameters for this module file that will be incorporated by the main file logging parameters
logger = logging.getLogger(__name__)

class CoordinationNumber(object):
    """
    Main class to compute CoordinationNumber
    """

    def __init__(self):
        """default constructor"""
        self.data = pd.DataFrame({"Step": np.empty([0])})

    @classmethod
    def from_trajectory(cls, trajectory, nb_set_and_cutoff, delta_Step = 1, first_frame = 0, parallel = False):
        """
        constructor of CoordinationNumber class from an ase trajectory object
        Args:
            nb_set_and_cutoff: dict, keys are str indicating pair of neighbours, 
                values are cutoffs float, in Angstrom
            dr: float, in Angstrom
        """
        cn_class = cls() # initialize class
        step = amof.trajectory.construct_step(delta_Step=delta_Step, first_frame = first_frame, number_of_frames = len(trajectory))
        cn_class.compute_cn(trajectory, nb_set_and_cutoff, step, parallel)
        return cn_class # return class as it is a constructor

    def compute_cn(self, trajectory, nb_set_and_cutoff, step, parallel):
        """
        compute coordination from ase trajectory object
        """
        atomic_numbers_unique = list(set(trajectory[0].get_atomic_numbers()))
        N_species = len(atomic_numbers_unique) # number of different chemical species

        logger.info("Start computing coordination number for %s frames", len(trajectory))
        cutoff_dict = amatom.format_cutoff(nb_set_and_cutoff)

        def compute_cn_for_frame(atom, step):
            """
            compute coordination for ase atom object
            """
            # atoms = trajectory[i]
            dic = {'Step': step}

            nl = amatom.get_neighborlist(atom, cutoff_dict) # 92% of computation time is spent in ase neighbour list search (profiled on ZIF8_15glass01)
            atomic_numbers = atom.get_atomic_numbers()  
            for nb_set, cutoff in nb_set_and_cutoff.items():
                a, b = tuple(ase.data.atomic_numbers[i] for i in nb_set.split('-'))
                cn_list = []
                for i in range(atom.get_global_number_of_atoms()):
                    if atomic_numbers[i] == a:
                        cn_list.append(np.sum(atomic_numbers[nl[i]] == b))
                dic[nb_set] = np.mean(cn_list)
            return dic

        if parallel == False:
            list_of_dict = [compute_cn_for_frame(trajectory[i], step[i]) for i in range(len(trajectory))]
        else:
            num_cores = parallel if type(parallel) == int else max(joblib.cpu_count() // 2 - 2, 2) # heuristic for 40cores Xeon cpus
            list_of_dict = joblib.Parallel(n_jobs=num_cores)(joblib.delayed(compute_cn_for_frame)(trajectory[i], step[i]) for i in range(len(trajectory)))

        self.data = pd.DataFrame(list_of_dict)

    @classmethod
    def from_file(cls, filename):
        """
        constructor of cn class from msd file
        """
        cn_class = cls() # initialize class
        cn_class.read_cn_file(filename)
        return cn_class # return class as it is a constructor

    def read_cn_file(self, filename):
        """path_to_data: where the cn object is"""
        filename = amof.files.path.append_suffix(filename, 'cn')
        self.data = pd.read_feather(filename)

    def write_to_file(self, filename):
        filename = amof.files.path.append_suffix(filename, 'cn')
        self.data.to_feather(filename)



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.13.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Elastic constants and mechanical properties
#
# Using a toy example of cell information stored during a [CP2K](https://www.cp2k.org/) MD run, we'll extract the elastic constants using the strain-fluctuation method before computing the resulting mechanical properties using [ELATE](https://github.com/coudertlab/elate/).
#
# Please note that due to the really small size of the cell information provided, Elastic Constant are not converged and the results unphysical.

# %%
import numpy as np

# %% [markdown]
# ### Prepare cell information
# Cell information can be stored under a variety of format depending on the MD scheme used.
#
# Here we show the example of the cell information formatted as a CP2K output, as aMOF contains a function to read it straight away, but any format of cell is possible as long as it end up formatted as required by aMOF (see below).

# %%
from amof.files.cp2k import read_tabular
cell_output = read_tabular('files/toy_trajectory.cell')

# %%
cell_output = cell_output[0:10000]

# %%
cell_output[0:5]

# %%
# Format as list of (a_x    a_y   a_z      b_x     b_y    b_z   c_x      c_y     c_z)
cell = np.array(cell_output.drop(columns=['Time', 'Volume']))
# Then reshape as list of ((a_x    a_y   a_z),
#                             (b_x     b_y    b_z),
#                               (c_x      c_y     c_z))
cell = cell.reshape(len(cell_output),3,3)

# %%
cell

# %% [markdown]
# ### Elastic Constants
# The module `elastic` contains the `ElasticConstant` class that can load the cell information under any format that ASE [.set_cell()](https://wiki.fysik.dtu.dk/ase/ase/atoms.html#ase.Atoms.set_cell) method can process

# %%
import amof.elastic as amela
help(amela.ElasticConstant.from_cell)

# %%
elastic_constant = amela.ElasticConstant.from_cell(cell, 300, final_value = True)

# %%
C = elastic_constant.Cmat
C

# %%
amela.print_Cmat(C)

# %% [markdown]
# ### Mechanical properties
# aMOF includes a wrapper to call [ELATE](https://github.com/coudertlab/elate/) to extract average mechanical properties through the `MechanicalProperties` class.

# %%
help(amela.MechanicalProperties.from_elastic)

# %%
C_as_list = np.array(C).tolist() # Convert to ELATE input
mech = amela.MechanicalProperties.from_elastic(C_as_list)

# %%
mech.data



"""
Module containing methods linked to elastic properties
"""

import math, sys
import numpy as np
import xarray as xr
import logging
import ase.atoms
import pandas as pd

import amof.files.path as ampath
import amof.elastic.elate as elate

logger = logging.getLogger(__name__)


class ElasticConstant(object):
    """
    Main class for computing elastic constants
    """

    def __init__(self):
        """default constructor, initialize with empty fields"""
        self.temperature = None
        self.h = None
        self.step = None
        self.volume = None
        self.epsilons = None
        # empty data array with proper dims and coords
        self.Cmat = xr.DataArray(np.empty([0,6,6]), 
            coords = [('Step', np.empty([0], dtype='int64')), ('row', np.arange(1,7)), ('col', np.arange(1,7))], 
            name = 'elastic')

    @classmethod
    def from_cell(cls, h, temperature, final_value = False, step = None):
        """default constructor from cell
        Args:
            h: array_like, unit cell tensor, can be of any form that ase set_cell can process
            temperature: float, temperature in K
            final_value: Boolean, If True only one value of C will be computed
            step: Contains step information.
                can be array_like of same length than h containing the step information
        """
        elastic_class = cls()
        elastic_class.temperature = temperature
        elastic_class.set_h(h)
        elastic_class.set_step(step)
        elastic_class.set_volume()
        elastic_class.set_epsilons()
        if final_value:
            elastic_class.set_final_C() 
        else:
            elastic_class.set_every_C()
        return elastic_class 


    def set_h(self, h):
        """
        Args:
            h: array_like, unit cell tensor, can be of any form that ase set_cell can process
        Results:
            self.h: Format is: ( a_x    a_y   a_z      b_x     b_y    b_z   c_x      c_y     c_z )
        """
        new_h = []
        mock_atom = ase.Atoms()
        for c in h:
            mock_atom.set_cell(c)
            new_h.append(mock_atom.get_cell().array)
        h = np.array(new_h)
        self.h = h
        
    def set_step(self, step):
        if step is None:
            self.step = None
        else:
            self.step = np.array(step) 

    @staticmethod
    def cummean(a):
        """
        Return the cumulative mean of the elements along a given axis.
        Args:
            a : array_like        Input array.
        """
        return np.cumsum(a) / np.arange(1, len(a) + 1)

    def set_volume(self):
        self.volume = self.cummean(list(map(np.linalg.det, self.h)))

    def set_epsilons(self):
        # Unit cell averages
        def vector_angle(v1, v2):
            v1_u = v1 / np.linalg.norm(v1)
            v2_u = v2 / np.linalg.norm(v2)
            angle = np.arccos(np.dot(v1_u, v2_u))
            if math.isnan(angle):
                if np.dot(v1_u, v2_u) > 0:
                    return 0.0
                else:
                    return np.pi
            return angle

        def h2abc(h):
            return (np.linalg.norm(h[0]), np.linalg.norm(h[1]), np.linalg.norm(h[2]),
                        vector_angle(h[1],h[2]), vector_angle(h[2],h[0]), vector_angle(h[0],h[1]))

        abc = list(map(h2abc, self.h))

        # Calculating the strain matrices

        inv_reference = np.linalg.inv(self.h[0])
        inv_reference_t = inv_reference.transpose()

        def h2eps(h):
            return (np.dot(inv_reference_t, np.dot(h.transpose(), np.dot(h, inv_reference))) - np.identity(3)) / 2

        self.epsilons = np.array(list(map(h2eps, self.h)))

    def set_every_C(self):
        # Elastic constants
        factor = (self.volume * 1.e-30) / (1.3806488e-23 * self.temperature)
        CARTESIAN_TO_VOIGT = ((0, 0), (1, 1), (2, 2), (2, 1), (2, 0), (1, 0))
        VOIGT_FACTORS = (1, 1, 1, 2, 2, 2)

        Smat = np.zeros((len(factor),6,6))
        for i in range(6):
            a, b = CARTESIAN_TO_VOIGT[i]
            fi = ElasticConstant.cummean([ epsilon[a, b] for epsilon in self.epsilons ])
            for j in range(i+1):
                u, v = CARTESIAN_TO_VOIGT[j]
                fj = ElasticConstant.cummean([ epsilon[u, v] for epsilon in self.epsilons ])
                fij = ElasticConstant.cummean([ epsilon[a, b] * epsilon[u, v] for epsilon in self.epsilons ])
                Smat[:, i,j] = VOIGT_FACTORS[i] * VOIGT_FACTORS[j] * factor * (fij - fi * fj)

        for i in range(5):
            for j in range(i+1,6):
                Smat[:,i,j] = Smat[:,j,i]

        # remove non inversible elements
        is_inversible = np.linalg.cond(Smat) < 1/sys.float_info.epsilon
        Smat = Smat[is_inversible]
        if self.step is not None:
            self.step = self.step[is_inversible]
        # Smat = Smat[1:] 

        # And now the stiffness matrix (in GPa)
        Cmat = np.linalg.inv(Smat) / 1.e9
        
        # turn to xarray object
        da = xr.DataArray(Cmat, dims=("Step","col",'row'))
        da["col"] = np.arange(1,7)
        da["row"] = np.arange(1,7)
        da.name = 'elastic'
        if self.step is not None:
            da["Step"] = self.step
        self.Cmat = da

    def set_final_C(self):
        volume = self.volume[-1]

        # Elastic constants
        factor = (volume * 1.e-30) / (1.3806488e-23 * self.temperature)
        CARTESIAN_TO_VOIGT = ((0, 0), (1, 1), (2, 2), (2, 1), (2, 0), (1, 0))
        VOIGT_FACTORS = (1, 1, 1, 2, 2, 2)

        Smat = np.zeros((6,6))
        for i in range(6):
            a, b = CARTESIAN_TO_VOIGT[i]
            fi = np.mean([ epsilon[a, b] for epsilon in self.epsilons ])
            for j in range(i+1):
                u, v = CARTESIAN_TO_VOIGT[j]
                fj = np.mean([ epsilon[u, v] for epsilon in self.epsilons ])
                fij = np.mean([ epsilon[a, b] * epsilon[u, v] for epsilon in self.epsilons ])
                Smat[i,j] = VOIGT_FACTORS[i] * VOIGT_FACTORS[j] * factor * (fij - fi * fj)

        for i in range(5):
            for j in range(i+1,6):
                Smat[i][j] = Smat[j][i]

        # And now the stiffness matrix (in GPa)
        Cmat = np.linalg.inv(Smat) / 1.e9
            
        # turn to xarray object
        da = xr.DataArray(Cmat, dims=("col",'row'))
        da["col"] = np.arange(1,7)
        da["row"] = np.arange(1,7)
        da.name = 'elastic'
        if self.step is not None:
            da["Step"] = self.step
        self.Cmat = da

    def write(self, filename):
        path_to_output = ampath.append_suffix(filename, 'elastic')
        self.Cmat.to_netcdf(path_to_output)


    @classmethod
    def from_file(cls, filename):
        """
        constructor of elastic file class from rdf file
        """
        new_class = cls() # initialize class
        new_class.read_elastic_file(filename)
        return new_class # return class as it is a constructor

    def read_elastic_file(self, filename):
        filename = ampath.append_suffix(filename, 'elastic')
        self.Cmat = xr.open_dataset(filename)



    # # not adapted
    # @staticmethod
    # def print_cell(abc):
    #     print('Unit cell averages:')
    #     print('            a = %.3f' % np.mean([x[0] for x in abc]))
    #     print('            b = %.3f' % np.mean([x[1] for x in abc]))
    #     print('            c = %.3f' % np.mean([x[2] for x in abc]))
    #     print('    alpha = %.3f' % np.rad2deg(np.mean([x[3] for x in abc])))
    #     print('        beta = %.3f' % np.rad2deg(np.mean([x[4] for x in abc])))
    #     print('    gamma = %.3f' % np.rad2deg(np.mean([x[5] for x in abc])))
    #     print('   volume = %.1f' % volume)


class MechanicalProperties(object):
    """
    Main class to compute Mechanical Properties from Elastic constants
    Wrapper of ELATE
    """

    def __init__(self):
        """default constructor"""
        self.data = pd.DataFrame()

    @classmethod
    def from_elastic(cls, C):
        """
        constructor of CoordinationNumber class from elastic constant tensor
        Args:
            C: list of list of float, elastic constant tensor
        """
        new_class = cls() # initialize class
        new_class.compute_averages(C)
        return new_class # return class as it is a constructor

    def compute_averages(self, C):
        """
        compute average mechanical properties using ELATE

        C: list of list of float, elastic constant tensor
        """
        el = elate.Elastic(C)
        prop = el.averages()
        df = pd.DataFrame(prop, 
            index = ["voigt", "reuss", "hill"], 
            columns = ["bulk_modulus","youngs_modulus","shear_modulus","poissons_ratio"])
        df.index.name = "averaging_scheme"
        self.data = df

    @classmethod
    def from_file(cls, filename):
        """
        constructor of cn class from msd file
        """
        new_class = cls() # initialize class
        new_class.read_file(filename)
        return new_class # return class as it is a constructor

    def read_file(self, filename):
        """path_to_data: where the cn object is"""
        filename = ampath.append_suffix(filename, 'mech.csv')
        self.data = pd.read_csv(filename, index_col=0)

    def write(self, filename):
        filename = ampath.append_suffix(filename, 'mech.csv')
        self.data.to_csv(filename)



def print_Cmat(Cmat):
    print('')
    print('Stiffness matrix C (GPa):')
    for i in range(6):
        print('    ', end=' ')
        for j in range(6):
            if j >= i:
                    print(('% 8.2f' % Cmat[i,j]), end=' ')
            else:
                    print('        ', end=' ')
        print('')

    # Eigenvalues
    print('')
    print('Stiffness matrix eigenvalues (GPa):')
    print((6*'% 8.2f') % tuple(np.sort(np.linalg.eigvals(Cmat))))



"""
Classes for nodes and linkers representation
"""

class BuildingUnit(object):
    """
    Abstract class for Building Units objects: nodes and linkers
    """
    def __init__(self) -> None:
        super().__init__()

class Node(BuildingUnit):
    """
    Generic class for representing nodes
    """
    def __init__(self) -> None:
        super().__init__()

class SingleMetal(Node):
    """
    Class for representing single metal nodes
    """
    def __init__(self, metal, target_coordination) -> None:
        """
        Args:
            metal: str, metal species (e.g. "Zn")
            target_coordination: int
        """
        super().__init__()
        self.name = metal
        self.species = [metal]
        self.target_coordination = target_coordination

class Linker(BuildingUnit):
    """
    Generic class for representing linkers
    """
    def __init__(self) -> None:
        super().__init__()

class ImidazoleBased(Linker):
    """
    Class for representing imidazole-based linkers
    """
    def __init__(self, name, formula) -> None:
        """
        Args:
            name: str, name of linker
            formula: str, chemical formula (e.g. "C3H3N2")
        """        
        super().__init__()
        self.name = name
        self.formula = formula
        self.species = ["C", "H", "N"] #temporary fix, to be changed (smiles & co)



"""
Instantiation of coordination search for ZIF glasses
Supported and validated: ZIF-4
"""

from numpy import False_
from amof.coordination import *
import amof.coordination.buildingunits as bu


class ZifSearch(CoordinationSearch):
    """
    Generic coordination search class for ZIFs comprised of: 
        single metal nodes
        imid-based linkers
    Provides generic methods used for all these systems
    """

    def __init__(self, struct, dist_margin=1.2, 
        dist_margin_metal=1.5, dist_margin_H=1.44,
        cutoff_metal = None, ignore_H_in_reduction = True):
        """
        Constructor of ZifSearch

        Args:
            dist_margin; float, default tolerance when using the covalent radii criteria 
                to determine if two atoms are neighbours
                The default option is equivalent to cutoffs of 1.752 for C-C and  1.728 for C-N bonds
            dist_margin_metal: float, specific tolerance for metal-X bonds as they're not covalent
                For Zn-N, the default option is equivalent to a cut-off of (1.22 + 0.71) * 1.5 = 2.895
            dist_margin_H: float, specific tolerance for H-X bonds as the distance covalent radius of H is small,
                causing the detection to be rather sensitive                
            cutoff_metal: float, override dist_margin_metal if not None 
                option added but not used so far
            ignore_H_in_reduction: bool, if True imid identification is only done by finding the cycle
                The errors in bonding H atoms are however reported, and the resulting reduced structure 
                    is computed with the H taken into account (barycenter).
        """
        self.dist_margin_metal = dist_margin_metal  # dist_margin is created in CoordinationSearch.__init__
        self.dist_margin_H = dist_margin_H  
        self.ignore_H_in_reduction = ignore_H_in_reduction
        neighb_max_distance = self.find_neighb_max_distance(
            dist_margin, dist_margin_metal, dist_margin_H, cutoff_metal)
        CoordinationSearch.__init__(
            self, struct, neighb_max_distance, dist_margin)
        self.detect_conn()
        self.clean_fragments()
        self.update_atypes()

    def find_neighb_max_distance(self, dist_margin, dist_margin_metal, dist_margin_H, cutoff_metal):
        """
        Return neighb_max_distance to compute all_neighb to the minimal necessary distance
        
        Args:
            dist_margin: float, default tolerance when using the covalent radii criteria to determine if two atoms are neighbours
            dist_margin_metal: float, specific tolerance for metal-X bonds
            dist_margin_H: float, specific tolerance for H-X bonds
            cutoff_metal: float, override dist_margin_metal if not None 
        """
        dist_margin_atoms = self.linker.species
        dist_margin_metal_atoms = self.node.species
        max_cov_linker = np.max([self.covalentradius[A]+self.covalentradius[B]
                                 for A in dist_margin_atoms for B in dist_margin_atoms])
        max_cov_H = np.max([self.covalentradius['H']+self.covalentradius[B]
                                for B in (dist_margin_atoms+dist_margin_metal_atoms)])                                 
        if cutoff_metal is None:
            max_cov_metal = np.max([self.covalentradius[A]+self.covalentradius[B]
                                for A in dist_margin_metal_atoms for B in (dist_margin_atoms+dist_margin_metal_atoms)])
        else: 
            max_cov_metal = cutoff_metal
        return max(max_cov_linker * dist_margin, max_cov_metal * dist_margin_metal, max_cov_H * dist_margin_H)

    def find_ABAcycles(self, A, B, cycle_length, target_number_of_cycles, fragtype = None):
        """
        Find every cycle of the form ABA...BA of length cycle_length for species A and B
        Add the bonds thus found to self.conn
        A can be neighbour to A and B, and B only to A. 
        For imid, A is C and B is N
        If fragtype is not None, will create new fragments with this fragname

        Args:
            A, B: string, lowercase string indicating species of A and B (e.g. "h")
            cycle_length: int
            target_number_of_cycles: int
            fragtype: str
        """
        graph = StructureGraph.with_empty_graph(self.struct)
        self.add_ABbonds(graph, B, A)
        self.add_ABbonds(graph, A, A)

        pattern = [A] + [B, A] * int((cycle_length - 1) / 2)
        cycles = self.find_rings(graph, pattern=pattern, target_number_of_rings = target_number_of_cycles, 
            exit_if_too_many_rings=False, remove_overlapping_rings=True)

        # check sanity of found cycles
        report_entry_1 = 'Expected number of cycles'
        self.report_search[report_entry_1] = (
            len(cycles) == target_number_of_cycles)
        if not self.report_search[report_entry_1]:
            logger.debug("number of cycles incorrect")
            self.report_search['Number of missing cycles'] = target_number_of_cycles - len(cycles)

        report_entry_2 = 'Atoms appear only once in cycles'
        in_cycle = [False for i in range(self.struct.num_sites)]
        self.report_search[report_entry_2] = True
        for c in cycles:
            for a, b in c:
                self.conn[a].append(b)
                self.conn[b].append(a)
                if in_cycle[a] == True:
                    logger.debug("atom %s appears in more than one cycle", a)
                    self.report_search[report_entry_2] = False
                in_cycle[a] = True
        self.clean_conn()

        if fragtype is not None:
            for c in cycles:
                indices = list(set(itertools.chain.from_iterable(c)))
                self.create_fragment(fragtype, indices)
        
        self.report_search['Cycle search successful'] = self.report_search[report_entry_1] and self.report_search[report_entry_2]

class MetalmIm(ZifSearch):
    """
    Search coordination for ZIFs made of metallic node and mIm (methyldimidazolate, C4N2H5) ligands
    Supports: ZIF-8
    Principle:
        1. find every cycle of CNCNC
        2. add H based on cov radii to single C and C bonded to one N
        3. bind the remaining H (there should be non for the crystal)
        4. bind the CHn to CN2
        5. bind N and Zn
    """

    def __init__(self, struct, metal, dist_margin=1.2, dist_margin_metal=1.5, ignore_H_in_reduction = True):
        """
        Constructor for coordination search for MetalmIm glasses

        Args:
            struct: pymatgen structure
            metal: str, representing metal node
            dist_margin: float, default tolerance when using the covalent radii criteria to determine if two atoms are neighbours
            dist_margin_metal: float, specific tolerance for metal-X bonds as they're not covalent
            ignore_H_in_reduction: bool, if True imid identification is only done by finding the cycle
        """
        self.node = bu.SingleMetal(metal, 4)
        self.linker = bu.ImidazoleBased("mIm", "C4N2H5")
        ZifSearch.__init__(self, struct, dist_margin = dist_margin, 
            dist_margin_metal = dist_margin_metal, 
            ignore_H_in_reduction = ignore_H_in_reduction)
    
    def detect_conn(self):
        """
        Main function to detect connectivity
        """
        # Find imid cycles (C-N-C-N-C)
        self.find_ABAcycles("c", "n", cycle_length = 5, 
            target_number_of_cycles = self.elems.count("n") / 2, 
            fragtype = self.linker.name)

        # hard way to force the reduction to work by ignoring the failed imid search
        if not self.report_search['Cycle search successful']:
            raise SearchError('Cycle search failed', self.report_search)

        # Connect H
        H_perfectly_connected = True

        # add H based on cov radii to single C and C bonded to one N
        new_fragments_name = 'methyl'
        report_entry = "C atoms missing H neighbours"
        C_Nbonds = self.get_A_Bbonds("c", "n")

        logger.debug("number of N nn to C atoms")  # debug check
        for i in range(3):
            logger.debug("%s C atoms have %s N nn", C_Nbonds.count(i), i)

        self.assign_B_uniquely_to_A_N_coordinated(
            lambda i: (C_Nbonds[i] in [0, 1]), 
            lambda i: (self.elems[i] == "h"),
            3, 
            report_level = 'undercoordinated', report_entry=report_entry, 
            propagate_fragments = True, new_fragments_name = new_fragments_name)
        H_perfectly_connected = H_perfectly_connected and self.report_search[report_entry] == []

        # bind the remaining H (there should be non for the crystal)
        # create a new fragment called irregular_H that can propagate to the entire imid or to Zn
        H_Cbonds = self.get_A_Bbonds("h", "c")
        new_fragments_name = self.linker.name if self.ignore_H_in_reduction else 'irregular_H'
        report_entry = "H atoms not bonded to C"
        self.find_N_closest_cov_dist(
            lambda i: H_Cbonds[i] == 0, 
            lambda i: True, 
            1, 
            report_level='full', report_entry=report_entry, 
            propagate_fragments = True, new_fragments_name = new_fragments_name)
        H_perfectly_connected = H_perfectly_connected and self.report_search[report_entry] == []

        self.report_search['H perfectly connected'] = H_perfectly_connected

        # link C in cycles (bonded to 2 N) to C bonded to H
        self.find_N_closest_cov_dist(
            lambda i: C_Nbonds[i] == 0, 
            lambda i: C_Nbonds[i] == 2, 
            1, 
            report_level='undercoordinated', report_entry="C in CHn not bonded to any C in imid", 
            propagate_fragments = 'reverse')

        # link N to metal_atom with no constraint on the number of N to metal_atom
        metal_atom = self.node.name.lower()
        self.assign_B_uniquely_to_A_N_coordinated(
            lambda i: self.elems[i] == metal_atom, 
            lambda i: self.elems[i] == "n",
            self.node.target_coordination, 
            dist_margin=self.dist_margin_metal, report_level='undercoordinated', 
            report_entry=f"undercoordinated {self.node.name}", new_fragments_name = self.node.name)

    def is_reduced_structure_valid(self):
        """
        Returns True iff nothing else then Im and Zn are found
        """
        return len(self.symbols.from_name_to_symbol) == 2

class MetalIm(ZifSearch):
    """
    Search coordination for ZIFs made of metallic node and Im (imidazolate, C3N2H3) ligands
    Supports: ZIF-4, ZIF-zni, SALEM-2
    Principle:
        1. find every cycle of CNCNC
        2. add H based on cov radii to every C
        3. bind the remaining H (there should be non for the crystal)
        4. bind N and Zn
    """

    def __init__(self, struct, metal, dist_margin=1.2, 
        dist_margin_metal=1.5, 
        dist_margin_H=1.44,
        ignore_H_in_reduction = True):
        """
        Constructor for coordination search for MetalmIm glasses

        Args:
            struct: pymatgen structure
            metal: str, representing metal node
            dist_margin: float, default tolerance when using the covalent radii criteria to determine if two atoms are neighbours
            dist_margin_metal: float, specific tolerance for metal-X bonds as they're not covalent
            dist_margin_H: float, specific tolerance for H-X bonds as the distance covalent radius of H is small,
                causing the detection to be rather sensitive
            ignore_H_in_reduction: bool, if True imid identification is only done by finding the cycle

        """
        self.node = bu.SingleMetal(metal, 4)
        self.linker = bu.ImidazoleBased("Im", "C3N2H3")
        ZifSearch.__init__(self, struct, dist_margin = dist_margin, 
            dist_margin_metal = dist_margin_metal, 
            dist_margin_H = dist_margin_H,
            ignore_H_in_reduction = ignore_H_in_reduction)

    
    def detect_conn(self):
        """
        Main function to detect connectivity
        """
        # Find imid cycles (C-N-C-N-C)
        self.find_ABAcycles("c", "n", cycle_length = 5, target_number_of_cycles = self.elems.count("n") / 2,
            fragtype = self.linker.name)

        # hard way to force the reduction to work by ignoring the failed imid search: to be investigated
        if not self.report_search['Cycle search successful']:
            raise SearchError('Cycle search failed', self.report_search)

        # Connect H
        H_perfectly_connected = True
        
        # add H based on cov radii to every C
        new_fragments_name = self.linker.name if self.ignore_H_in_reduction else 'irregular_C'
        report_entry = "C atoms missing H neighbours"
        self.assign_B_uniquely_to_A_N_coordinated(
            lambda i: (self.elems[i] == "c"), 
            lambda i: (self.elems[i] == "h"),  
            3, 
            report_level = 'undercoordinated', report_entry = report_entry,
            propagate_fragments = True, new_fragments_name = new_fragments_name,
            dist_margin=self.dist_margin_H) 
        H_perfectly_connected = H_perfectly_connected and self.report_search[report_entry] == []

        # bind the remaining H (there should be non for the crystal)
        H_Cbonds = self.get_A_Bbonds("h", "c")
        new_fragments_name = self.linker.name if self.ignore_H_in_reduction else 'irregular_H'
        report_entry = "H atoms not bonded to C"
        self.find_N_closest_cov_dist(
            lambda i: H_Cbonds[i] == 0, 
            lambda i: True, 
            1, 
            report_level='full', report_entry= report_entry,
            propagate_fragments = True, new_fragments_name = new_fragments_name,
            dist_margin=self.dist_margin_H) 
        H_perfectly_connected = H_perfectly_connected and self.report_search[report_entry] == []

        self.report_search['H perfectly connected'] = H_perfectly_connected

        # link N to metal_atom with no constraint on the number of N to metal_atom
        metal_atom = self.node.name.lower()

        self.assign_B_uniquely_to_A_N_coordinated(
            lambda i: self.elems[i] == metal_atom, 
            lambda i: self.elems[i] == "n",
            self.node.target_coordination, 
            dist_margin=self.dist_margin_metal, report_level='undercoordinated',
            report_entry=f"undercoordinated {self.node.name}", new_fragments_name = self.node.name)            

    def is_reduced_structure_valid(self):
        """
        Returns True iff nothing else then Im and Zn are found
        """
        return len(self.symbols.from_name_to_symbol) == 2


"""
Module containing methods acting on LAMMPS output files
"""

import logging
import os

logger = logging.getLogger(__name__)

def remove_duplicate_timesteps(filename):
    """
    remove duplicate timesteps in lammps xyz output
    search for lines starting by 'Atoms' and will remove the line before (number of atoms) until the start of the next frame.
    """
    seen_lines = set()
    with open(filename, "r") as fr:
        # lines = f.readlines()
        with open(str(filename)+"_temp_rm_duplicates", "w") as fw:
            previous = None
            write_to_file = True
            for line in fr:
                if line[0:5]=='Atoms':
                    if line not in seen_lines:
                        write_to_file = True
                        seen_lines.add(line)
                    else:
                        logger.info("Removing duplicate %s", line.strip("\n").strip('Atoms.'))
                        write_to_file = False
                if write_to_file and previous is not None:
                    fw.write(previous)
                previous = line
            if write_to_file:
                fw.write(previous)
    os.remove(filename)
    os.rename(str(filename)+"_temp_rm_duplicates", filename)


"""
Module containing functions to reduce a trajectory to its fragments
"""

import pandas as pd
import logging
import joblib

from pymatgen.io.ase import AseAtomsAdaptor
from amof.coordination.core import SearchError

import amof.symbols
import amof.coordination.zif
import amof.trajectory

logger = logging.getLogger(__name__)

def reduce_trajectory(trajectory, mof, filename = None, dist_margin = 1.2, delta_Step = 1, 
        first_frame = 0, parallel = False, write_mfpx = False, **kwargs):
    """
    Convenient wrapper to reduce trajectory by specifying a specific mof
    For can handle 'ZIF-4', 'ZIF-8', 'ZIF-zni' and 'SALEM-2'
    Thoroughly tested only for 'ZIF-4'

    Args:
        trajectory: ase trajectory object
        mof: str, mof name
        filename: str, where to write output files  
            if None, avoid writing
        dist_margin: float, default tolerance when using the covalent radii criteria to determine if two atoms are neighbours. 
            Default: 1.2
        delta_Step: int, number of simulation steps between two frames
        write_mfpx: bool, if True will write an mfpx file per frame of the reduced trajectory
    """
    dist_margin_metal = kwargs.get('dist_margin_metal', 1.6) # 1.6 means 3.088 for Zn-N (one case of 2.92 was observed)
    dist_margin_H = kwargs.get('dist_margin_H', 1.44)

    if mof in ['ZIF-4', 'ZIF-zni', 'SALEM-2']:
        structure_reducer = lambda struct: amof.coordination.zif.MetalIm(struct, "Zn", 
            dist_margin=dist_margin, 
            dist_margin_metal=dist_margin_metal,
            dist_margin_H = dist_margin_H) 
        symbols = amof.symbols.DummySymbols(['Zn', 'Im']) 
    elif mof in  ['ZIF-8']:
        structure_reducer = lambda struct: amof.coordination.zif.MetalmIm(struct, "Zn", dist_margin=dist_margin)
        symbols = amof.symbols.DummySymbols(['Zn', 'mIm']) 
    else:
        structure_reducer = lambda struct: amof.coordination.NotImplementedSearch(mof)
        symbols = amof.symbols.DummySymbols()
        logger.warning(f'Structure search not implemented for {mof}')
    return reduce_trajectory_core(trajectory, structure_reducer, symbols, filename, 
        delta_Step = delta_Step, first_frame = first_frame, parallel = parallel, write_mfpx = write_mfpx)

def reduce_trajectory_core(trajectory, structure_reducer, symbols, filename = None, delta_Step = 1, 
            first_frame = 0, parallel = False, write_mfpx = False):
        """
        Args:
            trajectory: ase trajectory object
            structure_reducer: a function taking a pymatgen structure as input and returning a CoordinationSearch object
            symbols: a DummySymbol object containing the representation to use for all frames 
                (no option atm to include additional symbols on specific frames,
                 parallelization issues + not adapted to analysis tools to have variable number of atoms)
            delta_Step: int, number of simulation steps between two frames
            filename: str, where to write output files, specify None to avoid writing
            write_mfpx: bool, if True will write an mfpx file per frame of the reduced trajectory
                if no filename is provided, will not write any
        
        Return:
            reduced_traj: amof.trajectory ReducedTrajectory object
        """
        logger.info("Start reducing trajectory for %s frames", len(trajectory))

        step = amof.trajectory.construct_step(delta_Step=delta_Step, first_frame = first_frame, number_of_frames = len(trajectory))

        def per_atom(atom, step, filename):
            report_search = {'Step': step}
            try:
                if filename is not None:
                    filename += f"_{step}" 

                # Non-wraped atoms were found to cause memory leak in CoordinationSearch at line
                # self.all_neighb = self.struct.get_all_neighbors(neighb_max_distance)          
                # It occured for extreme non wrapped atoms located ~50 images away, and didn't cause any issue for non-failing MD simulations
                atom.wrap() 

                reduced_atom, report_search_atom = reduce_atom(atom, structure_reducer, symbols, write_mfpx = write_mfpx, filename = filename)
                report_search['in_reduced_trajectory'] = reduced_atom is not None
                report_search = {**report_search, **report_search_atom}
            except SearchError as e:
                logger.debug('Failed to do reduce frame with error message: ' + e.message)        
                report_search['in_reduced_trajectory'] = False       
                report_search = {**report_search, **e.report_search}
                report_search['Error_message'] = e.message
                reduced_atom = None
            except BaseException as e: # unexpected exception
                logger.debug('Failed to do reduce frame with error message: ' + str(e))        
                report_search['in_reduced_trajectory'] = False       
                report_search['Error_message'] = "Unexpected Base Exception: " + str(e)
                reduced_atom = None
            return reduced_atom, report_search

        if parallel == False:
            result_list = [per_atom(trajectory[i], step[i], filename) for i in range(len(trajectory))]
        else:
            num_cores = parallel if type(parallel) == int else max(joblib.cpu_count() // 2 - 2, 2) # heuristic for 40cores Xeon cpus
            result_list = joblib.Parallel(n_jobs=num_cores)(joblib.delayed(per_atom)(trajectory[i], step[i], filename) for i in range(len(trajectory)))

        list_report_search = []
        reduced_traj = []
        for reduced_atom, report_search in result_list:
            list_report_search.append(report_search)
            if report_search['in_reduced_trajectory'] == True:
                reduced_traj.append(reduced_atom)
        
        df_report_search = pd.DataFrame(list_report_search).set_index('Step')

        reduced_trajectory = amof.trajectory.ReducedTrajectory(reduced_traj, df_report_search, symbols)
        if filename is not None:
            reduced_trajectory.write_to_file(filename)
        return reduced_trajectory

def reduce_atom(atom, structure_reducer, symbols, write_mfpx = False, filename = None):
    """
    Args:
        atom: ase atom object
        structure_reducer: a function taking a pymatgen structure as input and returning a CoordinationSearch object
        symbols: a DummySymbol object 
        write_mfpx: bool, if True will write an mfpx file per frame of the reduced trajectory
        filename: str, where to write output files. Shouldn't contain any extension
            If None doesn't write

    Return:
        reduced_atom: ase atom object with fragments instead of atoms
        report_search: dictionary
    """
    struct = AseAtomsAdaptor.get_structure(atom)
    searcher = structure_reducer(struct)
    searcher.symbols = symbols # enforce symbols; TBD: add option to get around, 'auto' setup, etc. 
    reduced_struct = searcher.reduce_structure()
    report_search = {**{"is_reduced_structure_valid": searcher.is_reduced_structure_valid()}, **searcher.report_search}
    if searcher.is_reduced_structure_valid():
        reduced_atom = AseAtomsAdaptor.get_atoms(reduced_struct)
        if write_mfpx == True and filename is not None:
            searcher.write_mfpx(filename)
    else:
        reduced_atom = None
    return reduced_atom, report_search



"""
Module containing methods acting on CP2K output files
"""

import logging
import os
import re
import pandas as pd

logger = logging.getLogger(__name__)

def clean_xyz(filename):
    """
    Clean cp2k xyz output:
        - remove duplicate timesteps:
    Works for positional data (positions, velocities, forces, etc.)

    search for lines starting by 'Atoms' and will remove the line before (number of atoms) until the start of the next frame.
    """
    seen_steps = set()
    with open(filename, "r") as fr:
        # lines = f.readlines()
        with open(str(filename)+"_temp_rm_duplicates", "w") as fw:
            previous = None
            write_to_file = True
            for line in fr:
                if line[0:5]==' i = ':
                    step = int(re.search(' i = (.*), time =(.*)', line).group(1))
                    if step not in seen_steps:
                        write_to_file = True
                        seen_steps.add(step)
                    else:
                        logger.info("Removing duplicate %s", step)
                        write_to_file = False
                if write_to_file and previous is not None:
                    fw.write(previous)
                previous = line
            if write_to_file:
                fw.write(previous)
    os.remove(filename)
    os.rename(str(filename)+"_temp_rm_duplicates", filename)


def clean_tabular(filename):
    """
    Clean cp2k tabular output (one line per step)
        - remove duplicate timesteps:
        - remove duplicate headers
    Works for ener, cell, stress
    """
    seen_steps = set()
    with open(filename, "r") as fr:
        # lines = f.readlines()
        with open(str(filename)+"_temp_rm_duplicates", "w") as fw:
            fw.write(fr.readline()) # write header (first line of first file)
            write_to_file = True
            for line in fr:
                if line[0]=='#':
                    write_to_file = False
                else:
                    step = int(re.split('\ +', line)[1])
                    if step not in seen_steps:
                        write_to_file = True
                        seen_steps.add(step)
                    else:
                        logger.info("Removing duplicate %s", line.strip("\n").strip('Atoms.'))
                        write_to_file = False
                if write_to_file:
                    fw.write(line)
    os.remove(filename)
    os.rename(str(filename)+"_temp_rm_duplicates", filename)    


def read_tabular(filename, return_units = False):
    """
    Parses tabular file
    Works for ener, cell, stress

    Args:
        filename: str or pathlib.Path
        return_units: Bool, if true returns unit_dict

    Return:
        df, pandas dataframe representing the data contained in filename
        unit_dict, dict of strings with columns names as keys and units as values
    """
    with open(filename, "r") as fr:
        first_line = fr.readline().strip('\n')
    # first_line = first_line.strip('#').strip('\n')
    columns = re.split('\  +', first_line)[1:] # rm '#' in first position, single whitespaces are tolerated in column name
    names = []
    units = []
    for c in columns:
        if 'Step' in c:
            names.append('Step')
            units.append('')
        else:
            search = re.search('(.*)\[(.*)\]', c)
            names.append(search.group(1).strip('.').strip(' ')) #strip points (abbreviations, not v. practical) and whitespaces (dataframe handling)
            units.append(search.group(2))

    df = pd.read_table(filename, skiprows=1, names=names, sep=r"\s+")
    df = df.set_index('Step')
    if return_units == False:
        return df
    else:
        return df, dict(zip(names, units))


# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.13.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Identify building units of ZIF-4
#
# Using the example of a single frame trajectory of a ZIF-4 crystal unit cell, we'll explore the building units identification and subsequent ring statistics computation.
#
# The module is `coordination`

# %%
# Import hvplot for easy plotting
import hvplot.xarray # noqa
# import hvplot.pandas # noqa
import numpy as np

# %% [markdown]
# #### Import trajectory
# Read with ase

# %%
import ase.io
atom = ase.io.read('files/ZIF-4.xyz')
traj = [atom] # single atom trajectory

# %%
traj[0].get_global_number_of_atoms()

# %% [markdown]
# ### Reduced trajectory
# First we create an reduced trajectory by identifying the building units for each frame of the trajectory and saving only the positions of these buildings units.

# %%
import amof.coordination.reduce as amred
help(amred)

# %%
red_traj = amred.reduce_trajectory(traj, 'ZIF-4')

# %% [markdown]
# This lead to the creation of a `ReducedTrajectory` object which contains a `trajectory` attribute with the reduced trajectory.

# %%
red_traj.trajectory[0].get_global_number_of_atoms()

# %% [markdown]
# It also contains a `report_search` attribute with the log of how the search for building units went

# %%
red_traj.report_search

# %% [markdown]
# ### Ring analysis
# Ring statistics can be computed from a reduced trajectory, which allows the computation of rings statistics of the metal-ligand network.

# %%
import amof.ring as amring
ring = amring.Ring.from_reduced_trajectory(red_traj, max_search_depth = 16)

# %%
ring.data.ring.hvplot.bar(x = 'ring_size')

# %%



"""
Forked from pysimm.apps.zeoplusplus
Modified:
    - introduced wait() in subprocess handling to avoid infinite loop
    - print to logging
    - removed everything linked to pysimm to have an independent module
    - removed nanohub
"""

# ******************************************************************************
# pysimm.appps.zeoplusplus module
# ******************************************************************************
#
# api to zeoplusplus simulation code
#
# ******************************************************************************
# License
# ******************************************************************************
# The MIT License (MIT)
#
# Copyright (c) 2019 Ping Lin, Coray M. Colina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys, os
# from subprocess import Popen, PIPE
import subprocess
from time import strftime
import shlex
import logging

logger = logging.getLogger(__name__)

ZEOpp_EXEC = os.environ.get('ZEOpp_EXEC')

def network(s, **kwargs):
    """pysimm.apps.zeopp.network

    Perform 1. Pore diameters; 2. Channel identification and dimensionality; 3. Surface area;
            4. Accessible volume; 5. Pore size distribution calculation using zeo++ v2.2
            6. Probe-occupiable volume; 7. Stochastic ray tracing; 8. Blocking spheres;
            9. Distance grids; 10. Structure analysis

    Args:
        s: filename of file in CSSR | CUC | V1 | CIF format
        atype_name: True to use atom type as atom name (usually need radii and mass info), False to use atom element
        radii: file name that contain atom radii data (rad.rad)
        mass: file name that contain atom mass data (mass.mass)
        probe_radius: radius of a probe used in sampling of surface (1.2 A)
        chan_radius: radius of a probe used to determine accessibility of void space (1.2 A)
        num_samples: number of Monte Carlo samples per unit cell (50000)
        option to include in the simulation: set True to activate
            ha: default=True, for using high accuracy,
            res: default=False, for diameters of the largest included sphere, the largest free sphere and the largest included sphere along free sphere path
            chan: default=False, for channel systems characterized by dimensionality as well as Di, Df and Dif
            sa: default=False, for surface area accessible to a spherical probe, characterized by
                      accessible surface area (ASA) and non-accessible surface area (NASA)
            vol: default=False, for accessible volume (AV) and non-accessible volume (NAV)
            volpo: default=False, for accessible proce-occupiable volume (POAV) and non-accessible probe-occupiable volume (PONAV)
            psd: default=False, for the "deriviative distribution" (change of AV w.r.t probe size) reported in the histogram file with 1000 bins of size of 0.1 Ang
            ray_atom: default=False
            block: default=False
            extra: user provided options, such as -gridG, -gridBOV, -strinfo, -oms, etc.
            
    ZEOpp_EXEC: path to zeo++ executable (network)

    Returns:
        None
    """
    global ZEOpp_EXEC

    if ZEOpp_EXEC is None:
        logger.debug('No environment variable ''ZEOpp_EXEC'' that points to '
              'zeo++ executable (network) \n Using curie path')
        ZEOpp_EXEC = '/home/nicolas/Applications/zeo++-0.3/network'

    probe_radius = kwargs.get('probe_radius', 1.2)
    chan_radius = kwargs.get('chan_radius', 1.2)
    num_samples = kwargs.get('num_samples', 50000)
    atype_name = kwargs.get('atype_name', False)

    ha = kwargs.get('ha', True)
    res = kwargs.get('res', False)
    chan = kwargs.get('chan', False)
    sa = kwargs.get('sa', False)
    vol = kwargs.get('vol', False)
    psd = kwargs.get('psd', False)
    volpo = kwargs.get('volpo', False)
    ray_atom = kwargs.get('ray_atom', False)
    block = kwargs.get('block', False)
    extra = kwargs.get('extra')
 
    if isinstance(s, str):
        input_file = s
        
    args = ZEOpp_EXEC

    if 'radii' in kwargs.keys(): 
        args += ' -r ' + kwargs.get('radii')
    if 'mass' in kwargs.keys(): 
        args += ' -mass ' + kwargs.get('mass')

    if ha:
        args += ' -ha'
    if res:
        args += ' -res'
    if chan:
        args += ' -chan ' + str(probe_radius) 
    if sa:
        args += ' -sa ' + str(chan_radius) + ' ' + str(probe_radius) + ' ' + str(num_samples) 
    if vol:
        args += ' -vol ' + str(chan_radius) + ' ' + str(probe_radius) + ' ' + str(num_samples) 
    if psd:
        args += ' -psd ' + str(chan_radius) + ' ' + str(probe_radius) + ' ' + str(num_samples) 
    if volpo:
        args += ' -volpo ' + str(chan_radius) + ' ' + str(probe_radius) + ' ' + str(num_samples)
    if ray_atom:
        args += ' -ray_atom ' + str(chan_radius) + ' ' + str(probe_radius) + ' ' + str(num_samples)
    if block:
        args += ' -block ' + str(probe_radius) + ' ' + str(num_samples)
    if extra:
        args += ' ' + extra

    args += ' ' + input_file

    arg_list = shlex.split(args)

    logger.debug('%s: starting simulation using zeo++'
          % strftime('%H:%M:%S'))
          
    p = subprocess.Popen(arg_list, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE) 
    # for very large system, zeopp spend a lot of time in the djikstra network made of cavities finding out which ones are accessible
    # tested for 20k atoms, it sometimes exit within 20-30min but can try for 14+ hours (exited manually)
    try:
        p.communicate(timeout=7200)
        p.wait()
    except subprocess.TimeoutExpired:
        p.kill()
        raise # raise last encountered exception

    logger.debug('%s: zeo++ simulation successful'
          % strftime('%H:%M:%S'))





"""
Module containing everything related to ase atom objects: 
- wrapper for ase atom module
"""

import ase
import ase.atoms
import ase.neighborlist
import numpy as np

def get_density(atom):
    """
    return density from an ase atom object
    """
    conversion_factor = 1.66053906660 # convert from uma per angstrom to SI (kg/L)
    return conversion_factor * get_total_mass(atom) / atom.get_volume()

def get_number_density(atom):
    """
    return number density in Angstrom^-3 from an ase atom object
    """
    return len(atom) / atom.get_volume()


def get_total_mass(atom):
    return np.sum(atom.get_masses())


def select_species_positions(atom, atomic_number):
    """
    return position of atom of species defined by atomic_number
    Args:
        atom: ase atom object
        atomic_number: int, atom selection 

    Return:
        position array of a given species
    """
    if atomic_number is None:
        return atom.get_positions()
    else:
        return atom.get_positions()[atom.get_atomic_numbers()==atomic_number]

def get_atomic_numbers_unique(atom):
    """return list of atomic numbers present in atom"""
    return list(set(atom.get_atomic_numbers()))

def format_cutoff(nb_set_and_cutoff, format='ase', sort_pair = False):
    """
    Return cutoff formatted for different purposes

    Args:
        nb_set_and_cutoff: dict, keys are str indicating pair of neighbors, 
            values are cutoffs float, in Angstrom
        format: str, can be 'ase'
        sort_pair: Bool, if True will sort the tuple of the nb atomic numbers

    Return:
        cutoff_dict: list of cutoff in ase.neighborlist format if format=='ase'
    """
    if format == 'ase':
        cutoff_dict = {} # dict in ase.nl cutoff format
        elements_present_unique = []
        for nn_set, cutoff in nb_set_and_cutoff.items():
            xx = tuple(ase.data.atomic_numbers[i] for i in nn_set.split('-'))
            if sort_pair == True:
                xx = tuple(sorted(xx))
            cutoff_dict[xx] = cutoff
            elements_present_unique += [ase.data.atomic_numbers[i] for i in nn_set.split('-')]
        return cutoff_dict

def get_neighborlist(atom, cutoff_dict):
    """
    Return neighbor list using ase

    Args:
        atom: ase atom object
        cutoff_dict: list of cutoff in ase.neighborlist format
    """
    

    nl_i, nl_j = ase.neighborlist.neighbor_list('ij', atom, cutoff_dict)
    nl = [[] for i in range(atom.get_global_number_of_atoms())]
    for k in range(len(nl_i)):
        i, j = nl_i[k], nl_j[k]
        nl[i].append(j)
    return nl